import functools
from taf.exceptions import RepositoryNotCleanError
from taf.git import GitRepository


def check_if_clean(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        skip_check = kwargs.pop("skip_clean_check", False)
        if not skip_check:
            path = kwargs.get("path", None)
            if path is None:
                path = args[0]
            repo = GitRepository(path=path)
            if repo.something_to_commit():
                raise RepositoryNotCleanError(repo.name)

        # Call the original function
        return func(*args, **kwargs)

    return wrapper



from typing import List, Optional, Any, Tuple
import subprocess


class TAFError(Exception):
    pass

    def __str__(self):
        try:
            return self.message
        except AttributeError:
            return super().__str__()


class CloneRepoException(TAFError):
    def __init__(self, repo):
        self.message = (
            f"Cannot clone {repo.name} from any of the following URLs: {repo.urls}"
        )


class CommandValidationError(TAFError):
    pass


class FetchException(TAFError):
    def __init__(self, path: str):
        self.message = f"Cannot fetch changes. Repo: {path}"


class GitError(TAFError):
    def __init__(
        self,
        repo: Any,
        command: Optional[str] = None,
        error: Optional[subprocess.CalledProcessError] = None,
        message: Optional[str] = None,
    ):
        if message is None:
            if command is not None:
                message = f"error occurred while executing {command}"
                if error is not None:
                    message = f"{message}:\n{str(error.stdout)}"
            elif error is not None:
                message = str(error)
            else:
                message = "error occurred"
        self.message = f"{repo.log_prefix}{message}" if repo is not None else message
        self.repo = repo
        self.command = command
        self.error = error


class InvalidBranchError(TAFError):
    pass


class InvalidCommitError(TAFError):
    pass


class InvalidKeyError(TAFError):
    def __init__(self, metadata_role: Optional[str]):
        super().__init__(
            f"Cannot sign {metadata_role} metadata file with inserted key."
        )


class InvalidOrMissingMetadataError(TAFError):
    pass


class InvalidRepositoryError(TAFError):
    pass


class InvalidPINError(TAFError):
    pass


class PushFailedError(GitError):
    pass


class RemoveMetadataKeyThresholdError(TAFError):
    def __init__(self, threshold: int):
        super().__init__(
            f"Remaining key number must be greater or equal to threshold ({threshold})."
        )


class RepositoryInstantiationError(TAFError):
    def __init__(self, repo_path: str, message: str):
        super().__init__(f"Could not instantiate repository {repo_path}\n\n: {message}")
        self.repo_path = repo_path
        self.message = message


class RepositoryNotCleanError(TAFError):
    def __init__(self, repo_name: str):
        message = f"Repository {repo_name} has uncommitted changes. Commit and push or revert the changes and run the command again."
        super().__init__(message)
        self.message = message


class MultipleRepositoriesNotCleanError(TAFError):
    def __init__(
        self,
        dirty_index_repos: List[str],
        unpushed_commits_repos_and_branches: List[Tuple[str, str]],
    ):
        message = ""
        dirty_repo_list = ", ".join(dirty_index_repos)
        if len(dirty_index_repos) >= 1:
            message += f"Repositories {dirty_repo_list} have uncommitted changes. Commit and push or use --force to revert and run the command again."
        unpushed_repo_branches = ", ".join(
            [
                f"{repo}: ({branch})"
                for repo, branch in unpushed_commits_repos_and_branches
            ]
        )
        if len(unpushed_commits_repos_and_branches) >= 1:
            message += f"\nThe following {'repository has' if len(unpushed_commits_repos_and_branches) == 1 else 'repositories have'} unpushed commits on branches: {unpushed_repo_branches}. Push the commits and run the command again."
        super().__init__(message)
        self.message = message


class NoRemoteError(GitError):
    def __init__(self, repo):
        message = f"No remotes configured for repository {repo.name}"
        super().__init__(message)


class ScriptExecutionError(TAFError):
    def __init__(self, script: str, error_msg: str):
        message = (
            f"An error happened during execution of script {script}:\n\n: {error_msg}"
        )
        super().__init__(message)
        self.message = message
        self.script = script


class SignersNotLoaded(TAFError):
    def __init__(self, roles):
        message = f"Signers of roles {', '.join(roles)} not loaded."
        super().__init__(message)


class MetadataUpdateError(TAFError):
    def __init__(self, metadata_role: str, message: str):
        super().__init__(
            f"Error happened while updating {metadata_role} metadata role(s):\n\n{message}"
        )
        self.metadata_role = metadata_role
        self.message = message


class MissingInfoJsonError(TAFError):
    pass


class NothingToCommitError(GitError):
    pass


class RootMetadataUpdateError(MetadataUpdateError):
    def __init__(self, message: str):
        super().__init__("root", message)


class KeystoreError(TAFError):
    pass


class PINMissmatchError(Exception):
    pass


class RolesKeyDataConversionError(TAFError):
    def __init__(self, exceptions: List[Exception]):
        message = "\n".join([str(exception) for exception in exceptions])
        super().__init__(message)


class SnapshotMetadataUpdateError(MetadataUpdateError):
    def __init__(self, message: str):
        super().__init__("snapshot", message)


class SigningError(TAFError):
    pass


class TargetsError(TAFError):
    def __init__(self, message: str):
        self.message = message


class TargetsMetadataUpdateError(MetadataUpdateError):
    def __init__(self, message: str):
        super().__init__("targets", message)


class TimestampMetadataUpdateError(MetadataUpdateError):
    def __init__(self, message: str):
        super().__init__("timestamp", message)


class PygitError(TAFError):
    pass


class NoSpeculativeBranchError(TAFError):
    pass


class RepositoriesNotFoundError(TAFError):
    pass


class UpdateFailedError(TAFError):
    pass


class ValidationFailedError(TAFError):
    pass


class YubikeyError(TAFError):
    pass



from taf.log import taf_logger
from pathlib import Path
from typing import Optional, Union


def find_taf_directory(auth_repo_path: Union[Path, str]) -> Optional[Path]:
    """Look for the .taf directory within the archive root.

    Args:
        auth_repo_path (Path): The path to the authentication repository.

    Returns:
        Optional[Path]: The path to the .taf directory if found, otherwise None.
    """
    # Check the parent directory of the authentication repository
    current_dir = Path(auth_repo_path).absolute().parent
    while current_dir != current_dir.root:
        taf_directory = current_dir / ".taf"
        if taf_directory.exists() and taf_directory.is_dir():
            return taf_directory
        current_dir = current_dir.parent

    # If not found, check the archive root
    archive_root = Path(auth_repo_path).parent.parent
    current_dir = archive_root
    while current_dir != current_dir.root:
        taf_directory = current_dir / ".taf"
        if taf_directory.exists() and taf_directory.is_dir():
            return taf_directory
        current_dir = current_dir.parent

    taf_logger.debug(
        f"No .taf directory found starting from {Path(auth_repo_path).parent}"
    )
    return None


def find_keystore(path: Union[str, Path]) -> Optional[Path]:
    """Find keystore starting from the given path and traversing parent directories if needed."""
    taf_directory = find_taf_directory(Path(path))
    if taf_directory:
        keystore_path = taf_directory / "keystore"
        if keystore_path.exists() and keystore_path.is_dir():
            taf_logger.debug(f"Found keystore at {keystore_path}")
            return keystore_path
    taf_logger.debug(f"No keystore found starting from {path}")
    return None



import json
from logging import ERROR, INFO
import shutil
from typing import Optional
import click
from logdecorator import log_on_end, log_on_error, log_on_start
from taf.git import GitRepository
from taf.messages import git_commit_message
from taf.models.types import RolesKeysData
from taf.models.converter import from_dict

from pathlib import Path
from taf.api.roles import (
    _initialize_roles_and_keystore,
)
from taf.api.targets import list_targets, register_target_files

from taf.auth_repo import AuthenticationRepository
from taf.exceptions import TAFError
from taf.keys import load_sorted_keys_of_new_roles
import taf.repositoriesdb as repositoriesdb
from taf.api.utils._conf import find_keystore
from taf.tuf.repository import METADATA_DIRECTORY_NAME
from taf.utils import ensure_pre_push_hook
from taf.log import taf_logger


@log_on_start(
    INFO, "Creating a new authentication repository {path:s}", logger=taf_logger
)
@log_on_end(INFO, "Finished creating a new repository", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while creating a new repository: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def create_repository(
    path: str,
    keystore: Optional[str] = None,
    roles_key_infos: Optional[str] = None,
    commit: Optional[bool] = False,
    test: Optional[bool] = False,
) -> None:
    """
    Create a new authentication repository. Generate initial metadata files.
    If target files already exist, add corresponding targets information to
    targets metadata files.

    Arguments:
        path: Authentication repository's location.
        keystore: Location of the keystore files.
        roles_key_infos: Path to a json file which contains information about repository's roles and keys.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        test (optional): Specifies if the created repository is a test authentication repository.

    Side Effects:
        Creates a new authentication repository (initializes a new git repository and generates tuf metadata)

    Returns:
        None
    """
    if not _check_if_can_create_repository(Path(path)):
        return

    if not keystore and path is not None:
        keystore_path = find_keystore(path)
        if keystore_path is not None:
            keystore = str(keystore_path)
    roles_key_infos_dict, keystore, skip_prompt = _initialize_roles_and_keystore(
        roles_key_infos, keystore
    )

    roles_keys_data = from_dict(roles_key_infos_dict, RolesKeysData)
    auth_repo = AuthenticationRepository(path=path)
    signers, verification_keys = load_sorted_keys_of_new_roles(
        roles=roles_keys_data.roles,
        yubikeys_data=roles_keys_data.yubikeys,
        keystore=keystore,
        skip_prompt=skip_prompt,
        certs_dir=auth_repo.certs_dir,
    )
    if signers is None:
        return

    repository = AuthenticationRepository(path=path)
    repository.create(roles_keys_data, signers, verification_keys)
    if commit:
        auth_repo.init_repo()
        commit_msg = git_commit_message("create-repo")
        auth_repo.commit(commit_msg, ["metadata"])

    if test:
        auth_repo.targets_path.mkdir(exist_ok=True)
        test_auth_file = auth_repo.targets_path / auth_repo.TEST_REPO_FLAG_FILE
        test_auth_file.touch()

    register_target_files(
        path,
        keystore,
        roles_key_infos,
        commit=commit,
        auth_repo=auth_repo,
        update_snapshot_and_timestamp=True,
        no_commit_warning=True,
    )

    ensure_pre_push_hook(auth_repo.path)

    if not commit:
        print("\nPlease commit manually.\n")


def _check_if_can_create_repository(path: Path) -> bool:
    """
    Check if a new authentication repository can be created at the specified location.
    A repository can be created if there is not directory at the repository's location
    or if it does exists, is not the root of a git repository.

    Arguments:
        auth_repo: Authentication repository.

    Side Effects:
        None

    Returns:
        True if a new authentication repository can be created, False otherwise.
    """
    repo = GitRepository(path=path)
    if path.is_dir():
        # check if there is non-empty metadata directory
        metadata_dir = path / METADATA_DIRECTORY_NAME
        if metadata_dir.is_dir() and any(metadata_dir.iterdir()):
            if repo.is_git_repository:
                print(
                    f'"{path}" is a git repository containing the metadata directory. Generating new metadata files could make the repository invalid. Aborting.'
                )
                return False
            if not click.confirm(
                f'Metadata directory found inside "{path}". Recreate metadata files?'
            ):
                return False
            else:
                shutil.rmtree(metadata_dir)
    return True


def taf_status(path: str, library_dir: Optional[str] = None, indent: int = 0) -> None:
    """
    Prints a list of target repositories of an authentication repository, their states,
    and the dependencies of the authentication repository.

    Arguments:
        path: Authentication repository's location
        library_dir (optional): Path to the library's root directory. Determined based on the authentication repository's path if not provided.
        indent (optional): Indentation level for nested dependencies.

    Side Effects:
       None

    Returns:
        None
    """
    # Get the authentication repository status
    print()
    auth_repo = AuthenticationRepository(path=path)
    head_commit = auth_repo.head_commit_sha()
    if head_commit is None:
        print("Repository is empty")
        return

    # Print authentication repository status
    indent_str = " " * indent
    print(f"{indent_str}Authentication Repository: {auth_repo.path.resolve()}")
    print(f"{indent_str}Head Commit: {head_commit}")
    print(f"{indent_str}Bare: {auth_repo.is_bare_repository}")
    print(f"{indent_str}Up to Date: {auth_repo.synced_with_remote()}")
    print(f"{indent_str}Something to commit: {auth_repo.something_to_commit()}")
    print(f"{indent_str}Target Repositories Status:")
    # Call the list_targets function
    print(json.dumps(list_targets(path=path), indent=1))

    # Load dependencies using repositoriesdb.get_auth_repositories
    repositoriesdb.load_dependencies(auth_repo, library_dir=library_dir)
    dependencies = repositoriesdb.get_auth_repositories(auth_repo, head_commit)
    if dependencies:
        print(f"{indent_str}Dependencies:")
        for dep_repo in dependencies.values():
            print(f"{indent_str}- {dep_repo.name}")
            taf_status(str(dep_repo.path), library_dir, indent + 3)



import enum
from attrs import define, field
from typing import Dict


@define
class Update:
    changed: bool = field(default=False)
    event: str = field(default="")
    error_msg: str = field(default="")
    auth_repos: Dict = field(factory=dict)
    auth_repo_name: str = field(default="")


class UpdateType(enum.Enum):
    TEST = "test"
    OFFICIAL = "official"
    EITHER = "either"


class OperationType(enum.Enum):
    CLONE = 1
    UPDATE = 2
    CLONE_OR_UPDATE = 3



from contextlib import contextmanager
from pathlib import Path
from typing import Dict, List, Optional, Union

from taf.api.utils._conf import find_keystore
from taf.auth_repo import AuthenticationRepository
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import PushFailedError, TAFError
from taf.keys import load_signers
from taf.log import taf_logger
from taf.messages import git_commit_message
from taf.constants import METADATA_DIRECTORY_NAME


@contextmanager
def transactional_execution(auth_repo):
    initial_commit = auth_repo.head_commit_sha()
    try:
        yield
    except PushFailedError:
        pass
    except Exception:
        auth_repo.reset_to_commit(initial_commit, hard=True)
        raise


@contextmanager
def manage_repo_and_signers(
    auth_repo: AuthenticationRepository,
    roles: Optional[List[str]] = None,
    keystore: Optional[Union[str, Path]] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    prompt_for_keys: Optional[bool] = False,
    paths_to_reset_on_error: Optional[List[Union[str, Path]]] = None,
    load_roles: Optional[bool] = True,
    load_parents: Optional[bool] = False,
    load_snapshot_and_timestamp: Optional[bool] = True,
    commit: Optional[bool] = True,
    push: Optional[bool] = True,
    commit_key: Optional[str] = None,
    commit_msg: Optional[str] = None,
    no_commit_warning: Optional[bool] = True,
):
    """
    A context manager that loads all signers and adds them to the specified authentication repository's
    signers cache. This allows for the execution of other methods without having to update the
    signers cache manually. Optionally, at the end, the context manager commits and pushes all changes made
    to the authentication repository and handles cleanup in case of an error.

    Arguments:
        auth_repo (AuthenticationRepository): Already instantiated authentication repository.
        roles (Optional[List[str]]): List of roles that are expected to be updated.
        keystore (Optional[Union[str, Path]]): Path to the keystore containing signing keys.
        scheme (Optional[str]): The signature scheme.
        prompt_for_keys (Optional[bool]): If True, prompts for keys if not found. Defaults to False.
        paths_to_reset_on_error (Optional[List[Union[str, Path]]]): Paths to reset if an error occurs.
        load_roles (Optional[bool]): If True, loads signing keys of the roles specified using the argument of the same name.
        load_parents (Optional[bool]): If true, loads sining keys of the specified roles' parents.
        load_snapshot_and_timestamp (Optional[bool]): If True, loads snapshot and timestamp signing keys.
        commit (Optional[bool]): If True, commits changes to the repository.
        push (Optional[bool]): If True, pushes changes to the remote repository.
        commit_key (Optional[str]): Commit key from `messages.py`
        commit_msg (Optional[str]): The message to use for commits.
        no_commit_warning (Optional[bool]): If True, suppresses warnings when not committing.
    """
    try:
        roles_to_load = set()
        if roles:
            unique_roles = set(roles)
            if load_roles:
                roles_to_load.update(unique_roles)
            if load_parents:
                roles_to_load.update(auth_repo.find_parents_of_roles(unique_roles))
        if load_snapshot_and_timestamp:
            roles_to_load.add("snapshot")
            roles_to_load.add("timestamp")
        if roles_to_load:
            if not keystore:
                keystore_path = find_keystore(auth_repo.path)
            else:
                keystore_path = Path(keystore)
            loaded_yubikeys: Dict = {}
            for role in roles_to_load:
                if not auth_repo.check_if_keys_loaded(role):
                    keystore_signers, yubikey_signers = load_signers(
                        auth_repo,
                        role,
                        loaded_yubikeys=loaded_yubikeys,
                        keystore=keystore_path,
                        scheme=scheme,
                        prompt_for_keys=prompt_for_keys,
                    )
                    auth_repo.add_signers_to_cache({role: keystore_signers})
                    auth_repo.add_signers_to_cache({role: yubikey_signers})
        yield
        if commit and auth_repo.something_to_commit():
            if not commit_msg and commit_key:
                commit_msg = git_commit_message(commit_key)
            auth_repo.commit_and_push(commit_msg=commit_msg, push=push)
        elif not no_commit_warning:
            taf_logger.log("NOTICE", "\nPlease commit manually\n")

    except PushFailedError:
        raise
    except Exception as e:
        taf_logger.error(f"An error occurred: {e}")
        if not paths_to_reset_on_error:
            paths_to_reset_on_error = [METADATA_DIRECTORY_NAME]
        elif METADATA_DIRECTORY_NAME not in paths_to_reset_on_error:
            paths_to_reset_on_error.append(METADATA_DIRECTORY_NAME)

        if auth_repo.is_git_repository and paths_to_reset_on_error:
            # restore metadata, leave targets as they might have been modified by the user
            # TODO flag for also resetting targets?
            # also update the CLI error handling]
            auth_repo.restore([str(path) for path in paths_to_reset_on_error])

        raise TAFError from e



import glob
import re
from getpass import getpass
from os import getcwd
from pathlib import Path
from typing import List, Optional
import click
import securesystemslib
from taf.tuf.keys import (
    load_public_key_from_file,
    load_signer_from_file,
    load_signer_from_pem,
)

from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import KeystoreError

from taf.tuf.repository import MetadataRepository as TUFRepository

from securesystemslib.signer._crypto_signer import CryptoSigner

from securesystemslib.signer._key import SSlibKey


def default_keystore_path() -> str:
    keystore_path = str(Path(getcwd(), "keystore"))
    return keystore_path


def get_keystore_keys_of_role(keystore: str, role: str) -> List[str]:
    keystores_glob = f"{keystore}/{role}*"
    # Use glob to find matching files
    keystores = glob.glob(keystores_glob)
    pattern = r"^{}(?:\d+)?$".format(re.escape(role))

    # Iterate over the matching files
    return [
        Path(keystore_path).name
        for keystore_path in keystores
        if re.match(pattern, Path(keystore_path).name)
    ]


def _form_private_pem(pem: str) -> str:
    return f"-----BEGIN RSA PRIVATE KEY-----\n{pem}\n-----END RSA PRIVATE KEY-----"


def _from_public_pem(pem: str) -> str:
    return f"-----BEGIN PUBLIC KEY-----\n{pem}\n-----END PUBLIC KEY-----"


def key_cmd_prompt(
    key_name: str,
    role: str,
    taf_repo: TUFRepository,
    loaded_keys: Optional[List] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
) -> CryptoSigner:
    def _enter_and_check_key(key_name, role, loaded_keys, scheme):
        pem = getpass(f"Enter {key_name} private key without its header and footer\n")
        pem = _form_private_pem(pem)
        try:
            signer = load_signer_from_pem(pem, scheme)
        except Exception:
            print("Invalid key")
            return None
        public_key = signer.public_key
        if not taf_repo.is_valid_metadata_yubikey(role, public_key):
            print(f"The entered key is not a valid {role} key")
            return None
        if loaded_keys is not None and public_key in loaded_keys:
            print("Key already entered")
            return None
        return signer

    while True:
        pem = _enter_and_check_key(key_name, role, loaded_keys, scheme)
        if pem is not None:
            return pem


def new_public_key_cmd_prompt(scheme: Optional[str]) -> SSlibKey:
    def _enter_and_check_key(scheme):
        pem = getpass("Enter public key without its header and footer\n")
        pem = _from_public_pem(pem)
        try:
            return load_public_key_from_file(pem, scheme)
        except Exception:
            print("Invalid key")
            return None

    while True:
        key = _enter_and_check_key(scheme)
        if key is not None:
            return key


def load_signer_from_private_keystore(
    keystore: str,
    key_name: str,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    password: Optional[str] = None,
) -> CryptoSigner:
    key_path = Path(keystore, key_name).expanduser().resolve()
    if not key_path.is_file():
        raise KeystoreError(f"{str(key_path)} does not exist")

    def _read_key(path, password, scheme):
        def _read_key_or_keystore_error(path, password, scheme):
            try:
                return load_signer_from_file(path, password or None, scheme=scheme)
            except TypeError:
                raise
            except Exception as e:
                raise KeystoreError(e)

        try:
            # try to load with a given password or None
            return _read_key_or_keystore_error(path, password, scheme)
        except TypeError:
            password = getpass(
                f"Enter {key_name} keystore file password and press ENTER"
            )
            try:
                return _read_key_or_keystore_error(path, password, scheme)
            except Exception:
                return None
        except Exception:
            return None

    while True:
        signer = _read_key(key_path, password, scheme)
        if signer is not None:
            return signer
        if not click.confirm(f"Could not open keystore file {key_path}. Try again?"):
            raise KeystoreError(f"Could not open keystore file {key_path}")


def read_public_key_from_keystore(
    keystore: str, key_name: str, scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME
) -> SSlibKey:
    pub_key_path = Path(keystore, f"{key_name}.pub").expanduser().resolve()
    if not pub_key_path.is_file():
        raise KeystoreError(f"{str(pub_key_path)} does not exist")
    try:
        return load_public_key_from_file(str(pub_key_path), scheme)
    except (
        securesystemslib.exceptions.FormatError,
        securesystemslib.exceptions.Error,
    ) as e:
        raise KeystoreError(e)



from typing import Dict
import pygit2
from collections import defaultdict
from taf.log import taf_logger
from taf.exceptions import GitError
import os.path


class PyGitRepository:
    def __init__(
        self,
        encapsulating_repo,
        *args,
        **kwargs,
    ):
        self.encapsulating_repo = encapsulating_repo
        self.path = encapsulating_repo.path
        self.repo = pygit2.Repository(str(self.path))

    _files_cache: Dict = defaultdict(dict)

    def _get_child(self, parent, path_part):
        """
        Return the child object of a parent object.
        Used for walking a git tree.
        """
        try:
            out = parent[path_part]
        except KeyError:
            return None
        else:
            return self.repo[out.id]

    def _get_object_at_path(self, obj, path):
        """
        for the given commit object,
        get the object at the given path
        """
        working = obj.tree
        if path.endswith("/"):
            path = path[:-1]
        path = path.split("/")
        for part in path:
            working = self._get_child(working, part)
            if working is None:
                return None
        return working

    def _get_blob_at_path(self, obj, path):
        """
        for the given commit object,
        get the blob at the given path
        """
        taf_logger.debug(
            f"Get blob at path {path}",
        )
        working = self._get_object_at_path(obj, path)
        if working and isinstance(working, pygit2.Blob):
            taf_logger.debug(f"Found blob at path {'/'.join(path)}")
            return working
        taf_logger.debug(f"Blob not found at path {'/'.join(path)}")
        return None

    def cleanup(self):
        """
        Must call this function in order to release pygit2 file handles.
        """
        self.repo.free()

    def get_file(self, commit, path, raw=False):
        """
        for the given commit string,
        return the string contents of the blob at the
        given path, if it exists, otherwise raise GitError
        """
        obj = self.repo.get(commit)
        blob = self._get_blob_at_path(obj, path)
        if blob is None:
            raise GitError(
                self.encapsulating_repo,
                message=f"fatal: Path '{path}' does not exist in '{commit}'",
            )
        else:
            git_id = blob.hex
            type = "raw" if raw else "decoded"
            if git_id not in self._files_cache or type not in self._files_cache[git_id]:
                content = blob.read_raw() if raw else blob.read_raw().decode()
                self._files_cache[git_id] |= {type: content}
            return git_id, self._files_cache[git_id][type]

    def _list_files_at_revision(self, tree, path="", results=None):
        """
        recurse through tree and return paths relative to that tree for
        all blobs in that tree.
        """
        if results is None:
            results = []

        for entry in tree:
            new_path = os.path.join(path, entry.name)
            if entry.type_str == "blob":
                results.append(new_path)
            elif entry.type_str == "tree":
                obj = self._get_child(tree, entry.name)
                self._list_files_at_revision(obj, new_path, results)
            else:
                raise NotImplementedError(
                    f"object at '{new_path}' of type '{entry.name}' not supported"
                )
        return results

    def list_files_at_revision(self, commit, path):
        """
        for the given commit string,
        return a list of all file paths that are
        descendents of the path string.
        """
        obj = self.repo.get(commit)
        root = self._get_object_at_path(obj, path)
        if root is None:
            raise GitError(
                self.encapsulating_repo,
                message=f"fatal: Path '{path}' does not exist in '{commit}'",
            )
        return self._list_files_at_revision(root)



import json
from pathlib import Path
import sys
import click
from taf.api.targets import (
    list_targets,
    add_target_repo,
    register_target_files,
    remove_target_repo,
    export_targets_history,
    update_and_sign_targets,
    update_target_repos_from_repositories_json
)
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import TAFError
from taf.tools.cli import catch_cli_exception, find_repository
from taf.log import taf_logger


def add_repo_command():
    @click.command(context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ), help="""Add a new repository by adding it to repositories.json, creating a delegation (if targets is not
        its signing role) and adding and signing initial target files if the repository is found on the filesystem.
        All additional information that should be saved as the repository's custom content in `repositories.json`
        is specified by providing a json file containing this data. If a new role should be added, this configuration
        file should also contain informatoin about that role.
        E.g.

        {
            "custom-prop1": "custom-val1",
            "custom-prop2": "custom-val2",
            "role": {
                "parent_role": "targets",
                "delegated_path": ["/delegated_path_inside_targets1", "/delegated_path_inside_targets2"],
                "keys_number": 1,
                "threshold": 1,
                "yubikey": true,
                "scheme": "rsa-pkcs1v15-sha256"
            }
        }

        parent_role = config_data.get("parent_role", "targets")
        keys_number = config_data.get("keys_number", 1)
        threshold = config_data.get("threshold", 1)
        yubikey = config_data.get("yubikey", False)
        scheme = config_data.get("scheme", DEFAULT_RSA_SIGNATURE_SCHEME)

        if directly inside the authentication repository.

        In this case, custom-prop1 and custom-prop2 will be added to the custom part of the target repository's entry in
        repositories.json.

        If the repository does not exist, it is sufficient to provide its namespace prefixed name
        instead of the full filesystem path. If the repository's path is not provided, it is expected
        to be located in the same library root directory as the authentication repository,
        in a directory whose name corresponds to its name. If authentication repository's path
        is `E:\\examples\\root\\namespace\\auth`, and the target's namespace prefixed name is
        `namespace1\\repo1`, the target's path will be set to `E:\\examples\\root\\namespace1\\repo1`.""")
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.argument("target-name")
    @click.option("--target-path", default=None, help="Target repository's filesystem path")
    @click.option("--role", default="targets", help="Signing role of the corresponding target file. Can be a new role, in which case it will be necessary to provide additional information")
    @click.option("--config-file", type=click.Path(exists=True), help="Path to the JSON configuration file containing information about the new role and/or targets custom data.")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not located inside the keystore directory")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be committed automatically")
    def add_repo(path, target_path, target_name, role, config_file, keystore, prompt_for_keys, scheme, no_commit):

        config_data = {}
        if config_file:
            try:
                config_data = json.loads(Path(config_file).read_text())
            except json.JSONDecodeError:
                click.echo("Invalid JSON provided. Please check your input.", err=True)
                sys.exit(1)

        if "role" in config_data:
            role_data = config_data.pop("role")
            delegated_path = role_data.get("delegated_path", [target_name])
            parent_role = role_data.get("parent_role", "targets")
            keys_number = role_data.get("keys_number", 1)
            threshold = role_data.get("threshold", 1)
            yubikey = role_data.get("yubikey", False)

            add_target_repo(
                path=path,
                target_path=target_path,
                target_name=target_name,
                library_dir=None,
                role=role,
                parent_role=parent_role,
                paths=delegated_path,
                keys_number=keys_number,
                threshold=threshold,
                yubikey=yubikey,
                keystore=keystore,
                custom=config_data,
                scheme=scheme,
                prompt_for_keys=prompt_for_keys,
                commit=not no_commit,
                should_create_new_role=True,
            )
        else:
            add_target_repo(
                path=path,
                target_path=target_path,
                target_name=target_name,
                library_dir=None,
                role=role,
                keystore=keystore,
                custom=config_data,
                scheme=scheme,
                prompt_for_keys=prompt_for_keys,
                commit=not no_commit,
                should_create_new_role=False,
            )
    return add_repo


def export_history_command():
    @click.command(help="""Export lists of sorted commits, grouped by branches and target repositories, based
        on target files stored in the authentication repository. If commit is specified,
        only return changes made at that revision and all subsequent revisions. If it is not,
        start from the initial authentication repository commit.
        Repositories which will be taken into consideration when collecting targets historical
        data can be defined using the repo option. If no repositories are passed in, historical
        data will include all target repositories.
        to a file whose location is specified using the output option, or print it to
        console.""")
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--commit", default=None, help="Starting authentication repository commit")
    @click.option("--output", default=None, help="File to which the resulting json will be written. If not provided, the output will be printed to console")
    @click.option("--repo", multiple=True, help="Target repository whose historical data should be collected")
    def export_history(path, commit, output, repo):
        export_targets_history(path, commit, output, repo)
    return export_history


def list_targets_command():
    @click.command(help="""List target repositories of the specified authentication repository. All target repositories
        are expected to be inside the same library root dir. Only repositories that are listed in
        repositories.json and whose corresponding target files exist (files whose name matched the
        name defined in repositories.json located inside the targets directory). For each repository,
        print:
        - if unauthenticated commits are allowed
        - if they exist on the user's local machine inside the library root dir (if they were cloned)
        - if they are bare
        - if there are unsigned changes (commits not registered in the authentication repository)
        - if they are up-to-date with remote
        - if there are uncommitted changes""")
    @find_repository
    @catch_cli_exception(handle=TAFError, print_error=True)
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    def list(path):
        targets_status = list_targets(path)
        taf_logger.log("NOTICE", json.dumps(targets_status, indent=4))
    return list


def remove_repo_command():
    @click.command(help="Remove a target repository (from repsoitories.json and target file) and sign")
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.argument("target-name")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not located inside the keystore directory")
    def remove_repo(path, target_name, keystore, prompt_for_keys):
        remove_target_repo(
            path=path,
            target_name=target_name,
            keystore=keystore,
            prompt_for_keys=prompt_for_keys,
        )
    return remove_repo


def sign_targets_command():
    @click.command(help="""Register and sign target files. This means that all targets metadata files corresponding
        to roles responsible for updated target files are updated. Once the targets
        files are updated, so are snapshot and timestamp. All files are then signed. If the
        keystore parameter is provided, keys stored in that directory will be used for
        signing. If a needed key is not in that directory, the file can either be signed
        by manually entering the key or by using a Yubikey.""")
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--keys-description", help="A dictionary containing information about the keys or a path to a json file which stores the needed information")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not located inside the keystore directory")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be committed automatically")
    def sign(path, keystore, keys_description, scheme, prompt_for_keys, no_commit):
        try:
            register_target_files(
                path=path,
                keystore=keystore,
                roles_key_infos=keys_description,
                scheme=scheme,
                update_snapshot_and_timestamp=True,
                prompt_for_keys=prompt_for_keys,
                commit=not no_commit
            )
        except TAFError as e:
            click.echo()
            click.echo(str(e))
            click.echo()
    return sign


def update_and_sign_command():
    @click.command(help="""Update target files corresponding to target repositories specified through the target type parameter
        by writing the current top commit and branch name to the target files. Sign the updated files
        and then commit. Types are expected to be defined in reposoitories.json, inside the custom data
        (Should be generalized in the future). If types are not specified, update all repositories specified
        in repositories.json.

        Target repositories are expected to be inside a directory whose name is equal to the specified
        namespace and which is located inside the root directory. If root directory is E:\\examples\\root
        and namespace is namespace1, target repositories should be in E:\\examples\\root\\namespace1.
        If the authentication repository and the target repositories are in the same root directory and
        the authentication repository is also directly inside a namespace directory, then the common root
        directory is calculated as two repositories up from the authentication repository's directory.
        Authentication repository's namespace can, but does not have to be equal to the namespace of target,
        repositories. If the authentication repository's path is E:\\root\\namespace\\auth-repo, root
        directory will be determined as E:\\root. If this default value is not correct, it can be redefined
        through the --library-dir option. If the --namespace option's value is not provided, it is assumed
        that the namespace of target repositories is equal to the authentication repository's namespace,
        determined based on the repository's path. E.g. Namespace of E:\\root\\namespace2\\auth-repo
        is namespace2.""")
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--library-dir", default=None, help="Directory where target repositories and, optionally, authentication repository are located. If omitted it is calculated based on authentication repository's path. Authentication repo is presumed to be at library-dir/namespace/auth-repo-name")
    @click.option("--target-type", multiple=True, help="Types of target repositories whose corresponding target files should be updated and signed. Should match a target type defined in repositories.json")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--keys-description", help="A dictionary containing information about the keys or a path to a json file which stores the needed information")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not located inside the keystore directory")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be committed automatically")
    def update_and_sign(path, library_dir, target_type, keystore, keys_description, scheme, prompt_for_keys, no_commit):
        try:
            if len(target_type):
                update_and_sign_targets(
                    path,
                    library_dir,
                    target_type,
                    keystore=keystore,
                    roles_key_infos=keys_description,
                    scheme=scheme,
                    prompt_for_keys=prompt_for_keys,
                    commit=not no_commit,
                )
            else:
                update_target_repos_from_repositories_json(
                    path,
                    library_dir,
                    add_branch=True,
                    keystore=keystore,
                    scheme=scheme,
                    prompt_for_keys=prompt_for_keys,
                    commit=not no_commit
                )
        except TAFError as e:
            click.echo()
            click.echo(str(e))
            click.echo()
    return update_and_sign


def attach_to_group(group):

    group.add_command(add_repo_command(), name='add-repo')
    group.add_command(export_history_command(), name='export-history')
    group.add_command(list_targets_command(), name='list')
    group.add_command(remove_repo_command(), name='remove-repo')
    group.add_command(sign_targets_command(), name='sign')
    group.add_command(update_and_sign_command(), name='update-and-sign')



import os
import sys
import logging
from typing import Dict
from pathlib import Path

# import tuf.log
# import tuf.exceptions
from loguru import logger as taf_logger
import taf.settings as settings

_CONSOLE_FORMAT_STRING = "\n{message}\n"
_FILE_FORMAT_STRING = "[{time}] [{level}] [{module}:{function}@{line}]\n{message}\n"

console_loggers: Dict = {}
file_loggers: Dict = {}

NOTICE = 25
taf_logger.level("NOTICE", no=NOTICE, color="<yellow>", icon="!")

VERBOSITY_LEVELS = {
    0: "NOTICE",  # Default
    1: "INFO",  # -v
    2: "DEBUG",  # -vv
    3: "TRACE",  # -vvv
}


def formatter(record):
    if record["level"].no == NOTICE:
        return f"<white>{_CONSOLE_FORMAT_STRING}</white>"
    elif record["level"].no == logging.WARNING:
        return f"<yellow>{_CONSOLE_FORMAT_STRING}</yellow>"
    elif record["level"].no == logging.INFO:
        return f"<blue>{_CONSOLE_FORMAT_STRING}</blue>"
    elif record["level"].no == logging.DEBUG:
        return f"<magenta>{_CONSOLE_FORMAT_STRING}</magenta>"
    elif record["level"].no == logging.ERROR:
        return f"<red>{_CONSOLE_FORMAT_STRING}</red>"
    else:
        return _CONSOLE_FORMAT_STRING


def disable_console_logging():
    try:
        taf_logger.remove(console_loggers["log"])
        # disable_tuf_console_logging()
    except ValueError:
        # will be raised if this is called twice
        pass


def disable_file_logging():
    try:
        taf_logger.remove(file_loggers["log"])
        # disable_tuf_console_logging()
    except ValueError:
        # will be raised if this is called twice
        pass


# def disable_tuf_console_logging():
#     try:
#         tuf.log.set_console_log_level(logging.CRITICAL)
#     except securesystemslib.exceptions.Error:
#         pass


# def disable_tuf_file_logging():
#     if tuf.log.file_handler is not None:
#         tuf.log.disable_file_logging()
#     else:
#         logging.getLogger("tuf").setLevel(logging.CRITICAL)
#     logging.getLogger("securesystemslib_keys").setLevel(logging.CRITICAL)
#     logging.getLogger("securesystemslib_util").setLevel(logging.CRITICAL)


def _get_log_location():
    location = settings.LOGS_LOCATION or os.environ.get("TAF_LOG")
    if location is None:
        location = Path.home() / ".taf"
        location.mkdir(exist_ok=True)
    else:
        location = Path(location)
    return location


def initialize_logger_handlers():
    taf_logger.remove()
    if settings.ENABLE_CONSOLE_LOGGING:
        console_loggers["log"] = taf_logger.add(
            sys.stdout, format=formatter, level=VERBOSITY_LEVELS[settings.VERBOSITY]
        )
        # tuf.log.set_console_log_level(logging.ERROR)
    # else:
    # if console logging is disable, remove tuf console logger
    # disable_tuf_console_logging()

    if settings.ENABLE_FILE_LOGGING:
        log_location = _get_log_location()
        log_path = str(log_location / settings.LOG_FILENAME)
        file_loggers["log"] = taf_logger.add(
            log_path, format=_FILE_FORMAT_STRING, level=settings.FILE_LOGGING_LEVEL
        )

        if settings.SEPARATE_ERRORS:
            error_log_path = str(log_location / settings.ERROR_LOG_FILENAME)
            file_loggers["error"] = taf_logger.add(
                error_log_path,
                format=_FILE_FORMAT_STRING,
                level=settings.ERROR_LOGGING_LEVEL,
            )
    #     try:
    #         tuf.log.set_filehandler_log_level(settings.FILE_LOGGING_LEVEL)
    #     except tuf.exceptions.Error:
    #         pass
    # else:
    # # if file logging is disabled, also disable tuf file logging
    # disable_tuf_file_logging()


initialize_logger_handlers()



import attrs
import datetime
from typing import List, Optional


TARGETS_DIRECTORY_NAME = "targets"
METADATA_DIRECTORY_NAME = "metadata"


DEFAULT_RSA_SIGNATURE_SCHEME = "rsa-pkcs1v15-sha256"

CAPSTONE = "capstone"
PROTECTED_DIRECTORY_NAME = "protected"
INFO_JSON_PATH = f"{TARGETS_DIRECTORY_NAME}/{PROTECTED_DIRECTORY_NAME}/info.json"


@attrs.define
class RoleSetupParams:
    number: int = attrs.field(default=1)
    threshold: int = attrs.field(default=1)
    yubikey: bool = attrs.field(default=False)
    scheme: str = attrs.field(default=DEFAULT_RSA_SIGNATURE_SCHEME)
    length: int = attrs.field(default=3072)
    passwords: Optional[List[str]] = attrs.field(default=None)
    terminating: bool = attrs.field(default=True)

    def __getitem__(self, key):
        return getattr(self, key, None)


DEFAULT_ROLE_SETUP_PARAMS = RoleSetupParams()

# Yubikey x509 certificate expiration interval
EXPIRATION_INTERVAL = 36500

YUBIKEY_EXPIRATION_DATE = datetime.datetime.now() + datetime.timedelta(
    days=EXPIRATION_INTERVAL
)



import json
from pathlib import Path
import sys
import click
from taf.api.roles import (
    add_multiple_roles,
    add_role,
    list_keys_of_role,
    add_signing_key,
    remove_role,
    revoke_signing_key,
    remove_paths,
    rotate_signing_key,
)
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import TAFError
from taf.auth_repo import AuthenticationRepository
from taf.log import taf_logger
from taf.tools.cli import catch_cli_exception, find_repository

from taf.api.roles import add_role_paths


def add_role_command():
    @click.command(help="""
    Add a new delegated target role. Allows optional specification of the role's properties through a JSON configuration file.
    If the configuration file is not provided or specific properties are omitted, default values are used.
    Only a list of one or more delegated paths has to be provided.

    Configuration file (JSON) can specify:
    - 'parent_role' (string): The parent role under which the new role will be delegated. Default is 'targets'.
    - 'delegated_path' (array of strings): Paths to be delegated to the new role. Must specify at least one if using a config file.
    - 'keys_number' (integer): Number of signing keys. Default is 1.
    - 'threshold' (integer): Number of keys required to sign. Default is 1.
    - 'yubikey' (boolean): Whether to use a YubiKey for signing. Default is false.
    - 'scheme' (string): Signature scheme, e.g., 'rsa-pkcs1v15-sha256'. Default is 'rsa-pkcs1v15-sha256'.

    Example JSON structure:
    {
    "parent_role": "targets",
    "delegated_path": ["/delegated_path_inside_targets1", "/delegated_path_inside_targets2"],
    "keys_number": 1,
    "threshold": 1,
    "yubikey": true,
    "scheme": "rsa-pkcs1v15-sha256"
    }
    """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.argument("role")
    @click.option("--config-file", type=click.Path(exists=True), help="Path to the JSON configuration file.")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be committed automatically")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not located inside the keystore directory")
    def add(role, config_file, path, keystore, no_commit, prompt_for_keys):

        config_data = {}
        if config_file:
            try:
                config_data = json.loads(Path(config_file).read_text())
            except json.JSONDecodeError:
                click.echo("Invalid JSON provided. Please check your input.", err=True)
                sys.exit(1)

        delegated_path = config_data.get("delegated_path", [])
        if not delegated_path:
            taf_logger.log("NOTICE", "Specify at least one delegated path through a configuration file.")
            return

        parent_role = config_data.get("parent_role", "targets")
        keys_number = config_data.get("keys_number", 1)
        threshold = config_data.get("threshold", 1)
        yubikey = config_data.get("yubikey", False)
        scheme = config_data.get("scheme", DEFAULT_RSA_SIGNATURE_SCHEME)

        add_role(
            path=path,
            role=role,
            parent_role=parent_role,
            paths=delegated_path,
            keys_number=keys_number,
            threshold=threshold,
            yubikey=yubikey,
            keystore=keystore,
            scheme=scheme,
            commit=not no_commit,
            prompt_for_keys=prompt_for_keys,
        )
    return add


def export_roles_description_command():
    @click.command(help="""
        Export roles-description.json file based on the
        """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--output", default=None, help="Output file path")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    def export_roles_description(path, output, keystore):
        auth_repo = AuthenticationRepository(path=path)
        roles_description = auth_repo.generate_roles_description()
        if keystore:
            roles_description["keystore"] = keystore
        if not output:
            taf_logger.log("NOTICE", json.dumps(roles_description, indent=True))
        else:
            output = Path(output)
            output.parent.mkdir(exist_ok=True, parents=True)
            output.write_text(json.dumps(roles_description, indent=True))

    return export_roles_description


def add_multiple_command():
    @click.command(help="""Adds new roles based on the provided keys-description file by
            comparing it with the current state of the repository.

            The current state can be exported using taf roles export_roles_description and then
            edited manually to add new roles.

            For each role, the following can be defined:
            - Total number of keys per role.
            - Threshold of required signatures per role.
            - Use of Yubikeys or keystore files for storing keys.
            - Signature scheme, with the default being 'rsa-pkcs1v15-sha256'.
            - Keystore path, if not specified via the keystore option.

        \b
        Example of a JSON configuration:
        {
            "roles": {
                "root": {
                    "number": 3,
                    "length": 2048,
                    "passwords": ["password1", "password2", "password3"],
                    "threshold": 2,
                    "yubikey": true
                },
                "targets": {
                    "length": 2048
                },
                "snapshot": {},
                "timestamp": {}
            },
            "keystore": "keystore_path"
        }
        """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.argument("keys-description")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be committed automatically")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not located inside the keystore directory")
    def add_multiple(path, keystore, keys_description, scheme, no_commit, prompt_for_keys):
        add_multiple_roles(
            path=path,
            keystore=keystore,
            roles_key_infos=keys_description,
            scheme=scheme,
            prompt_for_keys=prompt_for_keys,
            commit=not no_commit,
        )
    return add_multiple


def add_role_paths_command():
    @click.command(help="Add a new delegated target role, specifying which paths are delegated to the new role. Its parent role, number of signing keys and signatures threshold can also be defined. Update and sign all metadata files and commit.")
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.argument("role")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--delegated-path", multiple=True, help="Paths associated with the delegated role")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be committed automatically")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not located inside the keystore directory")
    def adding_role_paths(role, path, delegated_path, keystore, no_commit, prompt_for_keys):
        if not delegated_path:
            print("Specify at least one path")
            return

        add_role_paths(
            paths=delegated_path,
            delegated_role=role,
            keystore=keystore,
            commit=not no_commit,
            auth_path=path,
            prompt_for_keys=prompt_for_keys,
            push=not no_commit,
        )
    return adding_role_paths


# commenting out this command since its execution leads to an invalid state
# this is a TUF bug (or better said, caused by using a newer version of the updater and old repository_tool)
# it will be addressed when we transition to metadata API
def remove_role_command():
    @click.command(help="""Remove a delegated target role, and, optionally, its targets (depending on the remove-targets parameter).
        If targets should also be deleted, target files are remove and their corresponding entires are removed
        from repositoires.json. If targets should not get removed, the target files are signed using the
        removed role's parent role
        """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.argument("role")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--remove-targets/--no-remove-targets", default=True, help="Should targets delegated to this role also be removed. If not removed, they are signed by the parent role")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be committed automatically")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not located inside the keystore directory")
    def remove(role, path, keystore, scheme, remove_targets, no_commit, prompt_for_keys):
        remove_role(
            path=path,
            role=role,
            keystore=keystore,
            scheme=scheme,
            remove_targets=remove_targets,
            commit=not no_commit,
            prompt_for_keys=prompt_for_keys,
        )
    return remove


def remove_paths_command():
    @click.command(help="""Remove paths from delegated role""")
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--delegated-path", multiple=True, help="A list of paths to be removed")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be committed automatically")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not located inside the keystore directory")
    def remove_delegated_paths(path, delegated_path, keystore, scheme, no_commit, prompt_for_keys):
        if not delegated_path:
            print("Specify at least one role")
            return

        remove_paths(
            path=path,
            paths=delegated_path,
            keystore=keystore,
            scheme=scheme,
            commit=not no_commit,
            prompt_for_keys=prompt_for_keys,
        )
    return remove_delegated_paths


def add_signing_key_command():
    @click.command(help="""
        Add a new signing key. This will make it possible to a sign metadata files
        corresponding to the specified roles with another key. Although private keys are
        used for signing, key identifiers are calculated based on the public keys. This
        means that it's necessary to enter the public key in order to register
        a new signing key. Public key can be loaded from a file, in which case it is
        necessary to specify its path as the pub_key parameter's value. If this option
        is not used when calling this command, the key can be directly entered later.
        """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--role", multiple=True, help="A list of roles to whose list of signing keys the new key should be added")
    @click.option("--pub-key-path", default=None, help="Path to the public key corresponding to the private key which should be registered as the role's signing key")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be committed automatically")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not located inside the keystore directory")
    def adding_signing_key(path, role, pub_key_path, keystore, scheme, no_commit, prompt_for_keys):
        if not role:
            print("Specify at least one role")
            return

        add_signing_key(
            path=path,
            roles=role,
            pub_key_path=pub_key_path,
            keystore=keystore,
            scheme=scheme,
            commit=not no_commit,
            prompt_for_keys=prompt_for_keys
        )
    return adding_signing_key


def revoke_signing_key_command():
    @click.command(help="""
        Revoke a signing key.
        """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.argument("keyid")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--role", multiple=True, help="A list of roles from which to remove the key. If unspecified, the key is removed from all roles by default.")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be committed automatically")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not located inside the keystore directory")
    def revoke_key(path, role, keyid, keystore, scheme, no_commit, prompt_for_keys):

        revoke_signing_key(
            path=path,
            roles=role,
            key_id=keyid,
            keystore=keystore,
            scheme=scheme,
            commit=not no_commit,
            prompt_for_keys=prompt_for_keys
        )
    return revoke_key


def rotate_signing_key_command():
    @click.command(help="""
        Rotate a signing key.
        """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.argument("keyid")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--role", multiple=True, help="A list of roles from which to remove the key. Remove from all by default")
    @click.option("--pub-key-path", default=None, help="Path to the public key corresponding to the private key which should be registered as the role's signing key")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not located inside the keystore directory")
    @click.option("--revoke-commit-msg", default=None, help="Revoke key commit message")
    @click.option("--add-commit-msg", default=None, help="Add new signing key commit message")
    def rotate_key(path, role, keyid, pub_key_path, keystore, scheme, prompt_for_keys, revoke_commit_msg, add_commit_msg):
        rotate_signing_key(
            path=path,
            roles=role,
            key_id=keyid,
            keystore=keystore,
            scheme=scheme,
            prompt_for_keys=prompt_for_keys,
            pub_key_path=pub_key_path,
            revoke_commit_msg=revoke_commit_msg,
            add_commit_msg=add_commit_msg,
        )
    return rotate_key


def list_keys_command():
    @click.command(help="""
        List all keys of the specified role. If certs directory exists and contains certificates exported from YubiKeys,
        include additional information read from these certificates, like name or organization.
        """)
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.argument("role")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    def list_keys(role, path):
        key_infos = list_keys_of_role(
            path=path,
            role=role,
        )
        print("\n".join(key_infos))
    return list_keys


def attach_to_group(group):

    group.add_command(add_role_command(), name='add')
    group.add_command(add_multiple_command(), name='add-multiple')
    group.add_command(add_role_paths_command(), name='add-role-paths')
    group.add_command(remove_paths_command(), name='remove-paths')
    # group.add_command(remove_role_command(), name='remove')
    group.add_command(add_signing_key_command(), name='add-signing-key')
    group.add_command(revoke_signing_key_command(), name='revoke-key')
    group.add_command(rotate_signing_key_command(), name='rotate-key')
    group.add_command(list_keys_command(), name='list-keys')
    group.add_command(export_roles_description_command(), name="export-description")



from logging import DEBUG, ERROR
from typing import Dict, List, Optional, Tuple
import click
from collections import defaultdict
import json
from pathlib import Path
from logdecorator import log_on_end, log_on_error, log_on_start
from taf.api.api_workflow import manage_repo_and_signers, transactional_execution
from taf.tuf.keys import get_sslib_key_from_value
from taf.api.utils._git import check_if_clean
from taf.exceptions import KeystoreError, TAFError
from taf.models.converter import from_dict
from taf.models.types import TargetsRole, compare_roles_data
from taf.keys import (
    find_keystore,
    get_key_name,
    get_metadata_key_info,
    load_sorted_keys_of_new_roles,
)
from taf.auth_repo import AuthenticationRepository
from taf.constants import (
    DEFAULT_ROLE_SETUP_PARAMS,
    DEFAULT_RSA_SIGNATURE_SCHEME,
)
from taf.keystore import new_public_key_cmd_prompt
from taf.tuf.repository import MAIN_ROLES, METADATA_DIRECTORY_NAME
from taf.utils import get_key_size, read_input_dict, resolve_keystore_path
from taf.log import NOTICE, taf_logger
from taf.models.types import RolesKeysData
from taf.messages import git_commit_message

from securesystemslib.signer._key import SSlibKey


@log_on_start(DEBUG, "Adding a new role {role:s}", logger=taf_logger)
@log_on_end(DEBUG, "Finished adding a new role", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while adding a new role {role:s}: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
@check_if_clean
def add_role(
    path: str,
    role: str,
    parent_role: str,
    paths: list,
    keys_number: int,
    threshold: int,
    yubikey: bool,
    keystore: str,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    auth_repo: Optional[AuthenticationRepository] = None,
    commit: Optional[bool] = True,
    prompt_for_keys: Optional[bool] = False,
    push: Optional[bool] = True,
    skip_prompt: Optional[bool] = False,
) -> None:
    """
    Add a new delegated target role and update and sign metadata files.
    Automatically commit the changes if commit is set to True.

    Arguments:
        path: Path to the authentication repository.
        role: Name of the role which is to be added.
        parent_role: Name of the target role that is the new role's parent. Can be targets or another delegated role.
        paths: A list of target paths that are delegated to the new role.
        keys_number: Total number of signing keys
        threshold: Signature's threshold.
        yubikey: Specifies if the metadata file should be signed using a YubiKey.
        keystore: Location of the keystore files.
        scheme (optional): Signing scheme. Set to rsa-pkcs1v15-sha256 by default.
        auth_repo (optional): Instance of the authentication repository. Will be created if not passed into the function.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        push (optional): Flag specifying whether to push to remote.
        skip_prompt (optional): A flag defining if the user will be asked if they want to generate new keys or reuse existing
            ones in case keystore files should be used. New keys will be generated by default.

    Side Effects:
        Initializes a new delegated targets role, signs metadata files, write changes to the disk and optionally commits.

    Returns:
        None
    """

    if auth_repo is None:
        auth_repo = AuthenticationRepository(path=path)

    if not parent_role:
        parent_role = "targets"

    existing_roles = auth_repo.get_all_targets_roles()
    existing_roles.extend(MAIN_ROLES)
    if role in existing_roles:
        taf_logger.log("NOTICE", "All roles already set up")
        return

    keystore_path = Path(keystore) if keystore else find_keystore(Path(path))
    commit_msg = git_commit_message("add-role", role=role)
    metadata_path = Path(METADATA_DIRECTORY_NAME, f"{role}.json")

    with manage_repo_and_signers(
        auth_repo,
        roles=[parent_role],
        keystore=keystore_path,
        scheme=scheme,
        prompt_for_keys=prompt_for_keys,
        load_roles=True,
        load_snapshot_and_timestamp=True,
        commit=commit,
        push=push,
        commit_msg=commit_msg,
        paths_to_reset_on_error=[metadata_path],
    ):
        targets_parent_role = TargetsRole()
        if parent_role != "targets":
            targets_parent_role.name = parent_role
            targets_parent_role.paths = []

        new_role = TargetsRole()
        new_role.name = role
        new_role.parent = targets_parent_role
        new_role.paths = paths
        new_role.number = keys_number
        new_role.threshold = threshold
        new_role.yubikey = yubikey

        signers, _ = load_sorted_keys_of_new_roles(
            roles=new_role,
            yubikeys_data=None,
            keystore=keystore_path,
            skip_prompt=skip_prompt,
            certs_dir=auth_repo.certs_dir,
        )
        auth_repo.create_delegated_roles([new_role], signers)
        auth_repo.add_new_roles_to_snapshot([new_role.name])
        auth_repo.do_timestamp()


@log_on_start(DEBUG, "Adding new paths to role {delegated_role:s}", logger=taf_logger)
@log_on_end(DEBUG, "Finished adding new paths to role", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while adding new paths to role {delegated_role:s}: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def add_role_paths(
    paths: List[str],
    delegated_role: str,
    keystore: str,
    commit: Optional[bool] = True,
    auth_repo: Optional[AuthenticationRepository] = None,
    auth_path: Optional[str] = None,
    prompt_for_keys: Optional[bool] = False,
    push: Optional[bool] = True,
) -> None:
    """
    Adds additional delegated target paths to the specified role. That means that
    the role will be responsible for sining target files at those location going forward.

    Arguments:
        paths: A list of additional target paths that should be delegated to the role.
        delegated_role: Name of the targets role to which the new paths should be delegated.
        auth_path: Path to the authentication repository.
        keystore: Location of the keystore files.
        auth_repo (optional): Instance of the authentication repository. Will be created if not passed into the function.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        push (optional): Flag specifying whether to push to remote.
    Side Effects:
        Updates the specified target role's parent and other metadata files (snapshot and timestamp),
        signs them, writes changes to disk and optionally commits everything.

    Returns:
        None
    """

    if auth_repo is None:
        auth_repo = AuthenticationRepository(path=auth_path)

    parent_role = auth_repo.find_delegated_roles_parent(delegated_role)
    if all(
        path in auth_repo.get_delegations_of_role(parent_role)[delegated_role].paths
        for path in paths
    ):
        taf_logger.log("NOTICE", "Paths already added")
        return

    commit_msg = git_commit_message(
        "add-role-paths", role=delegated_role, paths=", ".join(paths)
    )

    with manage_repo_and_signers(
        auth_repo,
        [delegated_role],
        keystore=keystore,
        prompt_for_keys=prompt_for_keys,
        load_roles=False,
        load_parents=True,
        load_snapshot_and_timestamp=True,
        commit=commit,
        push=push,
        commit_msg=commit_msg,
    ):
        auth_repo.add_path_to_delegated_role(role=delegated_role, paths=paths)
        auth_repo.update_snapshot_and_timestamp()


@log_on_start(DEBUG, "Adding new roles", logger=taf_logger)
@log_on_end(DEBUG, "Finished adding new roles", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while adding new roles: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
@check_if_clean
def add_multiple_roles(
    path: str,
    keystore: Optional[str] = None,
    roles_key_infos: Optional[str] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    prompt_for_keys: Optional[bool] = False,
    commit: Optional[bool] = True,
    push: Optional[bool] = True,
) -> None:
    """
    Add new target roles and sign all metadata files given information stored in roles_key_infos
    dictionary or .json file.

    Arguments:
        path: Path to the authentication repository.
        keystore (optional): Location of the keystore files.
        roles_key_infos: Path to a json file which contains information about repository's roles and keys.
        scheme (optional): Signing scheme. Set to rsa-pkcs1v15-sha256 by default.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        push (optional): Flag specifying whether to push to remote.
    Side Effects:
        Updates metadata files (parent of new roles, snapshot and timestamp) and creates new targets metadata files.
        Writes changes to disk.

    Returns:
        None
    """

    roles_keys_data_new = _initialize_roles_and_keystore_for_existing_repo(
        path, roles_key_infos, keystore
    )

    auth_repo = AuthenticationRepository(path=path)
    roles_data = auth_repo.generate_roles_description()
    roles_keys_data_current = from_dict(roles_data, RolesKeysData)
    new_roles_data, _ = compare_roles_data(roles_keys_data_current, roles_keys_data_new)
    existing_roles = auth_repo.get_all_targets_roles()
    existing_roles.extend(MAIN_ROLES)
    roles_to_add_data = [
        role_data
        for role_data in new_roles_data
        if role_data.name not in existing_roles
    ]
    if not len(roles_to_add_data):
        taf_logger.log("NOTICE", "All roles already set up")
        return

    roles_to_add = [role_data.name for role_data in new_roles_data]
    commit_msg = git_commit_message("add-roles", roles=", ".join(roles_to_add))
    roles_to_load = [
        role_data.parent.name
        for role_data in new_roles_data
        if role_data.parent.name not in roles_to_add
    ]
    keystore_path = roles_keys_data_new.keystore

    with manage_repo_and_signers(
        auth_repo,
        roles=roles_to_load,
        keystore=keystore_path,
        scheme=scheme,
        prompt_for_keys=prompt_for_keys,
        load_snapshot_and_timestamp=True,
        commit_msg=commit_msg,
        commit=commit,
        push=push,
    ):
        all_signers = {}
        for role_to_add_data in roles_to_add_data:
            signers, _ = load_sorted_keys_of_new_roles(
                roles=role_to_add_data,
                yubikeys_data=None,
                keystore=keystore_path,
                skip_prompt=not prompt_for_keys,
                certs_dir=auth_repo.certs_dir,
            )
            all_signers.update(signers)

        auth_repo.create_delegated_roles(roles_to_add_data, all_signers)
        auth_repo.add_new_roles_to_snapshot(roles_to_add)
        auth_repo.do_timestamp()


@log_on_start(NOTICE, "Adding a new signing key", logger=taf_logger)
@log_on_end(DEBUG, "Finished adding a new signing key", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while adding a new signing key: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
@check_if_clean
def add_signing_key(
    path: str,
    roles: List[str],
    pub_key_path: Optional[str] = None,
    pub_key: Optional[SSlibKey] = None,
    keystore: Optional[str] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    commit: Optional[bool] = True,
    prompt_for_keys: Optional[bool] = False,
    push: Optional[bool] = True,
    commit_msg: Optional[str] = None,
) -> None:
    """
    Add a new signing key to the listed roles. Update root metadata if one or more roles is one of the main TUF roles,
    parent target role if one of the roles is a delegated target role and timestamp and snapshot in any case.

    Arguments:
        path: Path to the authentication repository.
        roles: A list of roles whose signing keys need to be extended.
        pub_key_path (optional): path to the file containing the public component of the new key. If not provided,
            it will be necessary to ender the key when prompted.
        keystore (optional): Location of the keystore files.
        scheme (optional): Signing scheme. Set to rsa-pkcs1v15-sha256 by default.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        push (optional): Flag specifying whether to push to remote.
        commit_msg(optional): Commit message. Will be necessary to enter it if not provided.
    Side Effects:
        Updates metadata files (parents of the affected roles, snapshot and timestamp).
        Writes changes to disk.

    Returns:
        None
    """

    pub_key = pub_key or _load_pub_key_from_file(
        pub_key_path, prompt_for_keys=prompt_for_keys, scheme=scheme
    )

    roles_keys = {role: [pub_key] for role in roles}

    auth_repo = AuthenticationRepository(path=path)

    with manage_repo_and_signers(
        auth_repo,
        roles,
        keystore,
        scheme,
        prompt_for_keys,
        load_snapshot_and_timestamp=True,
        load_parents=True,
        load_roles=False,
        commit=commit,
        push=push,
        commit_msg=commit_msg,
    ):
        added_keys, already_added_keys, invalid_keys = auth_repo.add_metadata_keys(
            roles_keys
        )
        if already_added_keys:
            taf_logger.log(
                "NOTICE", f"Key(s) {', '.join(already_added_keys)} already added"
            )
        if invalid_keys:
            taf_logger.warning(f"Key(s) {', '.join(invalid_keys)} invalid")

        if len(added_keys):
            auth_repo.update_snapshot_and_timestamp()


@log_on_start(NOTICE, "Revoking signing key", logger=taf_logger)
@log_on_end(DEBUG, "Finished revoking signing key", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while revoking signing key: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
@check_if_clean
def revoke_signing_key(
    path: str,
    key_id: str,
    roles: Optional[List[str]] = None,
    keystore: Optional[str] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    commit: Optional[bool] = True,
    prompt_for_keys: Optional[bool] = False,
    push: Optional[bool] = True,
    commit_msg: Optional[str] = None,
) -> None:
    """
    Revoke signing key. Update root metadata if one or more roles is one of the main TUF roles,
    parent target role if one of the roles is a delegated target role and timestamp and snapshot in any case.

    Arguments:
        path: Path to the authentication repository.
        roles: A list of roles whose signing keys need to be extended.
        key_id: id of the key to be removed
        keystore (optional): Location of the keystore files.
        scheme (optional): Signing scheme. Set to rsa-pkcs1v15-sha256 by default.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        push (optional): Flag specifying whether to push to remote.
        commit_msg(optional): Commit message. Will be necessary to enter it if not provided.
    Side Effects:
        Updates metadata files (parents of the affected roles, snapshot and timestamp).
        Writes changes to disk.

    Returns:
        None
    """

    auth_repo = AuthenticationRepository(path=path)

    roles_to_update = roles or auth_repo.find_keysid_roles([key_id])

    with manage_repo_and_signers(
        auth_repo,
        roles_to_update,
        keystore,
        scheme,
        prompt_for_keys,
        load_snapshot_and_timestamp=True,
        load_parents=True,
        load_roles=False,
        commit=commit,
        push=push,
        commit_msg=commit_msg,
    ):

        (
            removed_from_roles,
            not_added_roles,
            less_than_threshold_roles,
        ) = auth_repo.revoke_metadata_key(key_id=key_id, roles=roles)
        if not_added_roles:
            taf_logger.log(
                "NOTICE",
                f"Key is not a signing key of role(s) {', '.join(not_added_roles)}",
            )
        if less_than_threshold_roles:
            taf_logger.warning(
                f"Cannot remove key from {', '.join(less_than_threshold_roles)}. Number of keys must be greater or equal to thresholds"
            )

        if len(removed_from_roles):
            auth_repo.update_snapshot_and_timestamp()


@check_if_clean
def rotate_signing_key(
    path: str,
    key_id: str,
    pub_key_path: Optional[str] = None,
    roles: Optional[List[str]] = None,
    keystore: Optional[str] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    prompt_for_keys: Optional[bool] = False,
    push: Optional[bool] = True,
    revoke_commit_msg: Optional[str] = None,
    add_commit_msg: Optional[str] = None,
) -> None:
    """
    Rotate signing key. Remove it from one or more roles and add a new signing key.
    Update root metadata if one or more roles is one of the main TUF roles,
    parent target role if one of the roles is a delegated target role and timestamp and snapshot in any case.

    Arguments:
        path: Path to the authentication repository.
        roles: A list of roles whose signing keys need to be extended.
        key_id: id of the key to be removed
        pub_key_path (optional): path to the file containing the public component of the new key. If not provided,
            it will be necessary to ender the key when prompted.
        keystore (optional): Location of the keystore files.
        scheme (optional): Signing scheme. Set to rsa-pkcs1v15-sha256 by default.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        push (optional): Flag specifying whether to push to remote.
        revoke_commit_msg(optional): First commit message, when revokig the specified key. Will be necessary to enter it if not provided.
        add_commit_msg(optional): Second commit message, when addug a new signing key. Will be necessary to enter it if not provided.
    Side Effects:
        Updates metadata files (parents of the affected roles, snapshot and timestamp).
        Writes changes to disk.

    Returns:
        None
    """

    pub_key = _load_pub_key_from_file(
        pub_key_path, prompt_for_keys=prompt_for_keys, scheme=scheme
    )
    auth_repo = AuthenticationRepository(path=path)
    roles = roles or auth_repo.find_keysid_roles([key_id])

    with transactional_execution(auth_repo):
        revoke_signing_key(
            path=path,
            key_id=key_id,
            roles=roles,
            keystore=keystore,
            scheme=scheme,
            commit=True,
            prompt_for_keys=prompt_for_keys,
            push=False,
            commit_msg=revoke_commit_msg,
        )

        add_signing_key(
            path=path,
            roles=roles,
            pub_key=pub_key,
            keystore=keystore,
            scheme=scheme,
            commit=True,
            prompt_for_keys=prompt_for_keys,
            push=push,
            commit_msg=add_commit_msg,
        )


# TODO this is probably outdated, the format of the outputted roles_key_infos
def _enter_roles_infos(keystore: Optional[str], roles_key_infos: Optional[str]) -> Dict:
    """
    Ask the user to enter information taf roles and keys, including the location
    of keystore directory if not entered through an input parameter

    Arguments:
        keystore: Location of the keystore files.
        roles_key_infos: Path to a json file which contains information about repository's roles and keys.

    Side Effects:
        None

    Returns:
        A dictionary containing entered information about taf roles and keys (total number of keys per role,
        parent roles of roles, threshold of signatures per role, indicator if metadata should be signed using
        a yubikey for each role, key length and signing scheme for each role)
    """
    mandatory_roles = ["root", "targets", "snapshot", "timestamp"]
    role_key_infos: Dict = defaultdict(dict)
    infos_json: Dict = {}

    for role in mandatory_roles:
        role_key_infos[role] = _enter_role_info(role, role == "targets", keystore)
    infos_json["roles"] = role_key_infos

    if keystore:
        infos_json["keystore"] = keystore

    def _print_roles_key_infos(infos_json_str):
        print("------------------")
        print(
            "Configuration json - save it in order to make creation of repositories quicker"
        )
        print(json.dumps(infos_json, indent=4))
        print("------------------")

    infos_json_str = json.dumps(infos_json, indent=4)
    if roles_key_infos is not None:
        try:
            path = Path(roles_key_infos)
            path.parent.mkdir(parents=True, exist_ok=True)
            Path(roles_key_infos).write_text(infos_json_str)
            taf_logger.log(
                "NOTICE",
                f"Configuration json written to {Path(roles_key_infos).absolute()}",
            )
        except Exception as e:
            taf_logger.error(e)
            _print_roles_key_infos(infos_json_str)
    else:
        print(infos_json_str)
    return infos_json


def _enter_role_info(
    role: str, is_targets_role: bool, keystore: Optional[str] = None
) -> Dict:
    role_info = {}
    keys_num = _read_val(int, f"number of {role} keys", "number")
    if keys_num is not None:
        role_info["number"] = keys_num

    role_info["yubikey"] = click.confirm(f"Store {role} keys on Yubikeys?")
    if role_info["yubikey"]:
        # in case of yubikeys, length and scheme have to have specific values
        role_info["length"] = 2048
        role_info["scheme"] = DEFAULT_RSA_SIGNATURE_SCHEME
    else:
        # if keystore is specified and contains keys corresponding to this role
        # get key size based on the public key
        keystore_length = 0
        if keystore is not None:
            keystore_length = _get_roles_key_size(role, keystore, keys_num)
        if keystore_length == 0:
            key_length = _read_val(int, f"{role} key length", "length")
            if key_length is not None:
                role_info["length"] = key_length
        else:
            role_info["length"] = keystore_length
        scheme = _read_val(str, f"{role} signature scheme", "scheme")
        if scheme is not None:
            role_info["scheme"] = scheme

    threshold = _read_val(int, f"{role} signature threshold", "threshold")
    if threshold is not None:
        role_info["threshold"] = threshold

    if is_targets_role:
        delegated_roles: Dict = defaultdict(dict)
        while click.confirm(
            f"Add {'another' if len(delegated_roles) else 'a'} delegated targets role of role {role}?"
        ):
            role_name = _read_val(str, "role name", "role_name", True)
            delegated_paths: List[str] = []
            while not len(delegated_paths) or click.confirm("Enter another path?"):
                delegated_paths.append(
                    _read_val(
                        str,
                        f"path or glob pattern delegated to {role_name}",
                        "delegated_paths",
                        True,
                    )
                )
            delegated_roles[role_name]["paths"] = delegated_paths
            is_terminating = click.confirm(f"Is {role_name} terminating?")
            delegated_roles[role_name]["terminating"] = is_terminating
            delegated_roles[role_name].update(
                _enter_role_info(role_name, True, keystore)
            )
        role_info["delegations"] = delegated_roles

    return role_info


def _load_pub_key_from_file(pub_key_path, prompt_for_keys, scheme) -> SSlibKey:
    pub_key_pem = None
    if pub_key_path is not None:
        pub_key_pem_path = Path(pub_key_path)
        if pub_key_pem_path.is_file():
            pub_key_pem = Path(pub_key_path).read_text()

    if pub_key_pem is None and prompt_for_keys:
        pub_key_pem = new_public_key_cmd_prompt(scheme)["keyval"]["public"]

    if pub_key_pem is None:
        raise TAFError("Public key not provided or invalid")

    return get_sslib_key_from_value(pub_key_pem)


def _read_val(input_type, name, param=None, required=False):
    default_value_msg = ""
    default_value = None
    if param is not None:
        default_value = DEFAULT_ROLE_SETUP_PARAMS[param]
        if default_value is not None:
            default_value_msg = f"(default {default_value}) "

    while True:
        try:
            val = input(f"Enter {name} and press ENTER {default_value_msg}")
            if not val:
                if not required:
                    return default_value
                else:
                    continue
            return input_type(val)
        except ValueError:
            pass


def _initialize_roles_and_keystore_for_existing_repo(
    path: str,
    roles_key_infos: Optional[str],
    keystore: Optional[str],
    enter_info: Optional[bool] = True,
) -> RolesKeysData:
    roles_key_infos_dict = read_input_dict(roles_key_infos)

    if not roles_key_infos_dict and enter_info:
        roles_key_infos_dict = _enter_roles_infos(None, roles_key_infos)
    roles_keys_data = from_dict(roles_key_infos_dict, RolesKeysData)
    keystore = keystore or roles_keys_data.keystore
    if keystore is None and path is not None:
        keystore_path = find_keystore(Path(path))
        if keystore_path:
            roles_keys_data.keystore = str(keystore_path)
    else:
        roles_keys_data.keystore = keystore
    return roles_keys_data


def _initialize_roles_and_keystore(
    roles_key_infos: Optional[str],
    keystore: Optional[str],
    enter_info: Optional[bool] = True,
) -> Tuple[Dict, Optional[str], bool]:
    """
    Read information about roles and keys from a json file or ask the user to enter
    this information if not specified through a json file and enter_info is True.

    Arguments:
        roles_key_infos: A dictionary containing information about the roles:
            - total number of keys per role
            - their parent roles
            - threshold of signatures per role
            - should keys of a role be on Yubikeys or should a keystore files be used
            - scheme (the default scheme is rsa-pkcs1v15-sha256)
            - keystore path, if not specified via keystore option
        keystore: Location of the keystore files.
        enter_info (optional): Indicates if the user should be asked to enter information about the
        roles and keys if not specified. Set to True by default.


    Side Effects:
        None

    Returns:
        A dictionary containing entered information about taf roles and keys (total number of keys per role,
        parent roles of roles, threshold of signatures per role, indicator if metadata should be signed using
        a yubikey for each role, key length and signing scheme for each role) and keystore file path.
    """

    skip_prompt = False

    roles_key_infos_dict = read_input_dict(roles_key_infos)

    if not roles_key_infos_dict and enter_info:
        roles_key_infos_dict = _enter_roles_infos(None, roles_key_infos)

    # Check if all keys should be loaded from/stored to Yubikeys
    use_yubikeys = all(
        role_info.get("yubikey", False) for role_info in roles_key_infos_dict.values()
    )

    if not use_yubikeys and not keystore:
        keystore = roles_key_infos_dict.get("keystore")
        if not keystore:
            while True:
                use_keystore = (
                    input("Do you want to save/load keys from keystore files? [y/N]: ")
                    .strip()
                    .lower()
                )
                if use_keystore in ["y", "n"]:
                    break
            if use_keystore == "y":
                keystore = (
                    input("Enter keystore path (default ./keystore): ").strip()
                    or "./keystore"
                )
            else:
                taf_logger.log(
                    "NOTICE",
                    "Keys will be entered and then printed from the command line...",
                )

    if keystore is not None:
        keystore = resolve_keystore_path(keystore, roles_key_infos)
        roles_key_infos_dict["keystore"] = keystore

        while True:
            keystore_path = Path(keystore)
            if keystore_path.exists():
                break
            create_keystore = (
                input(
                    f"Keystore directory {keystore_path} does not exist. Do you want to create it? [y/N]: "
                )
                .strip()
                .lower()
            )
            if create_keystore == "y":
                keystore_path.mkdir(parents=True, exist_ok=True)
                roles_key_infos_dict["keystore"] = keystore
                print(f"Created keystore directory at {keystore}")
                skip_prompt = True
                break
            else:
                enter_new_path = (
                    input(
                        "Do you want to enter a different path to the keystore? [y/N]: "
                    )
                    .strip()
                    .lower()
                )
                if enter_new_path == "y":
                    keystore = input("New keystore path: ").strip()
                    keystore = resolve_keystore_path(keystore, roles_key_infos)
                    roles_key_infos_dict["keystore"] = keystore
                else:
                    raise KeystoreError("Keystore not found")

    return roles_key_infos_dict, keystore, skip_prompt


def _get_roles_key_size(role: str, keystore: str, keys_num: int) -> int:
    pub_key_name = f"{get_key_name(role, 1, keys_num)}.pub"
    key_path = str(Path(keystore, pub_key_name))
    return get_key_size(key_path)


@log_on_error(
    ERROR,
    "Could not list keys of {role:s}: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def list_keys_of_role(
    path: str,
    role: str,
) -> List[str]:
    """
     Print information about signing keys of role. If a certificate whose name matches
     a key's id exists, include information contained by that certificate (like name and valid to/from dates)

     Arguments:
         path: Path to the authentication repository.
         role: Name of the role which is to be removed.

    Side Effects:
         None

     Returns:
         None
    """
    auth_repo = AuthenticationRepository(path=path)
    key_ids = auth_repo.get_role_keys(role=role)
    if key_ids is None:
        raise TAFError(f"Role {role} does not exist")

    return [
        str(get_metadata_key_info(auth_repo.certs_dir, key_id)) for key_id in key_ids
    ]


@log_on_start(DEBUG, "Removing role {role:s}", logger=taf_logger)
@log_on_end(DEBUG, "Finished removing the role", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while removing role {role:s}: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
@check_if_clean
def remove_role(
    path: str,
    role: str,
    keystore: str,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    commit: Optional[bool] = True,
    remove_targets: Optional[bool] = False,
    auth_repo: Optional[AuthenticationRepository] = None,
    prompt_for_keys: Optional[bool] = False,
    push: Optional[bool] = True,
) -> None:
    """
    Remove a delegated target role and update and sign metadata files.
    Automatically commit the changes if commit is set to True.
    It is not possible to remove any of the main TUF roles

    Arguments:
        path: Path to the authentication repository.
        role: Name of the role which is to be removed.
        keystore: Location of the keystore files.
        scheme (optional): Signing scheme. Set to rsa-pkcs1v15-sha256 by default.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        remove_targets (optional): Indicates if target files should be removed to, or signed by the parent role.
            Set to False by default.
        auth_repo (optional): Instance of the authentication repository. Will be created if not passed into the function.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        push (optional): Flag specifying whether to push to remote.
    Side Effects:
        Updates metadata files, optionally deletes target files, writes changes to disk and optionally commits.

    Returns:
        None
    """

    # TODO This didn't fully work and was not being used
    # can be done later

    # if role in MAIN_ROLES:
    #     taf_logger.error(
    #         f"Cannot remove role {role}. It is one of the roles required by the TUF specification"
    #     )
    #     return

    # if auth_repo is None:
    #     auth_repo = AuthenticationRepository(path=path)

    # parent_role = auth_repo.find_delegated_roles_parent(role)
    # if parent_role is None:
    #     taf_logger.error("Role is not among delegated roles")
    #     return
    # parent_role_obj = auth_repo._role_obj(parent_role, auth_repo)
    # if not isinstance(parent_role_obj, Targets):
    #     taf_logger.error(f"Could not find parent targets role of role {role}.")
    #     return

    # roleinfo = tuf.roledb.get_roleinfo(parent_role, auth_repo.name)
    # added_targets_data: Dict = {}
    # removed_targets = []
    # for delegations_data in roleinfo["delegations"]["roles"]:
    #     if delegations_data["name"] == role:
    #         paths = delegations_data["paths"]
    #         for target_path in paths:
    #             target_file_path = Path(path, TARGETS_DIRECTORY_NAME, target_path)
    #             if target_file_path.is_file():
    #                 if remove_targets:
    #                     os.unlink(str(target_file_path))
    #                     removed_targets.append(str(target_file_path))
    #                 else:
    #                     added_targets_data[target_file_path] = {}
    #             else:
    #                 # try glob pattern traversal
    #                 full_pattern = str(Path(path, TARGETS_DIRECTORY_NAME, target_path))
    #                 matching_files = glob.glob(full_pattern)
    #                 for file_path in matching_files:
    #                     if remove_targets:
    #                         os.unlink(str(file_path))
    #                         removed_targets.append(file_path)
    #                     else:
    #                         added_targets_data[file_path] = {}
    #         break

    # parent_role_obj.revoke(role)

    # _update_role(
    #     auth_repo, role=parent_role, keystore=keystore, prompt_for_keys=prompt_for_keys
    # )
    # if len(added_targets_data):
    #     removed_targets_data: Dict = {}
    #     update_target_metadata(
    #         auth_repo,
    #         added_targets_data,
    #         removed_targets_data,
    #         keystore,
    #         write=False,
    #         scheme=DEFAULT_RSA_SIGNATURE_SCHEME,
    #         prompt_for_keys=prompt_for_keys,
    #     )

    # # if targets should be deleted, also removed them from repositories.json
    # if len(removed_targets):
    #     repositories_json = repositoriesdb.load_repositories_json(auth_repo)
    #     if repositories_json is not None:
    #         repositories = repositories_json["repositories"]
    #         for removed_target in removed_targets:
    #             if removed_target in repositories:
    #                 repositories.pop(removed_target)

    #             # update content of repositories.json before updating targets metadata
    #             Path(auth_repo.path, REPOSITORIES_JSON_PATH).write_text(
    #                 json.dumps(repositories_json, indent=4)
    #             )

    # update_snapshot_and_timestamp(
    #     auth_repo, keystore, scheme=scheme, prompt_for_keys=prompt_for_keys
    # )
    # if commit:
    #     commit_msg = git_commit_message("remove-role", role=role)
    #     auth_repo.commit_and_push(commit_msg=commit_msg, push=push)
    # else:
    #     taf_logger.log("NOTICE", "Please commit manually")


@log_on_start(DEBUG, "Removing delegated paths", logger=taf_logger)
@log_on_end(DEBUG, "Finished removing delegated paths", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while removing roles: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def remove_paths(
    path: str,
    paths: List[str],
    keystore: str,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    commit: Optional[bool] = True,
    commit_msg: Optional[str] = None,
    prompt_for_keys: Optional[bool] = False,
    push: Optional[bool] = True,
) -> bool:
    """
    Remove delegated paths. Update parent roles of the roles associated with the removed paths,
    as well as snapshot and timestamp. Optionally commit the changes.

    Arguments:
        path:  Path to the authentication repository.
        paths: Paths to be removed.
        keystore: Location of the keystore files.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        push (optional): Flag specifying whether to push to remote.
    Side Effects:
        Updates metadata files, writes changes to disk and optionally commits them.

    Returns:
        True if the delegation existed, False otherwise
    """
    auth_repo = AuthenticationRepository(path=path)
    paths_to_remove_from_roles = defaultdict(list)
    for path_to_remove in paths:
        delegated_role = auth_repo.get_role_from_target_paths([path_to_remove])

        if delegated_role != "targets":
            paths_to_remove_from_roles[delegated_role].append(path_to_remove)
        else:
            taf_logger.log("NOTICE", f"Path {path_to_remove} not delegated to any role")
    if not len(paths_to_remove_from_roles):
        taf_logger.log("NOTICE", "No paths delegated")
        return False

    if commit_msg is None:
        commit_msg = git_commit_message(
            "remove-role-paths", paths=", ".join(paths), role=delegated_role
        )
    with manage_repo_and_signers(
        auth_repo,
        roles=list(paths_to_remove_from_roles.keys()),
        keystore=keystore,
        scheme=scheme,
        prompt_for_keys=prompt_for_keys,
        load_roles=False,
        load_parents=True,
        load_snapshot_and_timestamp=True,
        commit=commit,
        push=push,
        commit_msg=commit_msg,
    ):
        auth_repo.remove_delegated_paths(paths_to_remove_from_roles)
        auth_repo.update_snapshot_and_timestamp()

    return True


def list_roles(auth_repo: AuthenticationRepository) -> None:
    """
    Print a list of all defined roles with their thresholds and parent roles.
    """

    def print_role_and_children(role: str, indent: int = 0) -> None:
        threshold = auth_repo.get_role_threshold(role)
        indent_str = " " * indent
        print(f"{indent_str}{role} (threshold: {threshold})")

        # Retrieve children (delegated roles)
        delegations = auth_repo.get_delegations_info(role)
        if delegations:
            children = [role_info["name"] for role_info in delegations.get("roles", [])]
            for child in children:
                print_role_and_children(child, indent + 2)

    for role in MAIN_ROLES:
        print_role_and_children(role)



import shutil
import sys
import click
from functools import partial, wraps
from logging import ERROR
from logdecorator import log_on_error

from taf.exceptions import InvalidRepositoryError, TAFError
from taf.log import taf_logger
from taf.repository_utils import find_valid_repository
from taf.git import GitRepository
from taf.utils import is_run_from_python_executable, on_rm_error


def catch_cli_exception(func=None, *, handle=TAFError, print_error=False, remove_dir_on_error=False, skip_cleanup=False):
    if not func:
        return partial(
            catch_cli_exception,
            handle=handle,
            print_error=print_error,
            remove_dir_on_error=remove_dir_on_error,
            skip_cleanup=skip_cleanup,
        )

    handle = tuple(handle) if isinstance(handle, (list, tuple)) else (handle,)

    @wraps(func)
    def wrapper(*args, **kwargs):
        successful = False
        try:
            result = func(*args, **kwargs)
            successful = True
            return result
        except handle as e:
            # TODO
            # for now
            # if print_error:
            taf_logger.error(e)
        except Exception as e:
            if is_run_from_python_executable():
                taf_logger.error(f"An error occurred: {e}")
                sys.exit(1)
            else:
                raise e
        finally:
            if not skip_cleanup and not successful and "path" in kwargs:
                path = kwargs["path"]
                repo = GitRepository(path=path)
                if repo.is_git_repository and not repo.is_bare_repository:
                    repo.restore(["metadata"])
                    # repo.clean_and_reset()

                if remove_dir_on_error:
                    shutil.rmtree(path, onerror=on_rm_error)

    return wrapper


def find_repository(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            if "path" in kwargs:
                kwargs["path"] = find_valid_repository(kwargs["path"])
            return func(*args, **kwargs)
        except InvalidRepositoryError as e:
            click.echo(f"An error occurred: {e}")
            sys.exit(1)

    return wrapper


@log_on_error(
    ERROR,
    "{e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
def process_custom_command_line_args(ctx):
    """
    Certain cli commands need to make it possible to enter custom arguments.
    E.g. when adding a new target repository, users can specify using the cli
    key-value pairs that will be written to its custom section of repositories.json.
    This function checks if that input is valid. It is easy to accidentally omit a space,
    e.g. --arg1 val1--arg2 val2. So show a nice error message in such a case. Also convert
    boolean values to actual booleans.

    Arguments:
        ctx: click context

    Returns:
       A dictionary containing parsed custom arguments
    """

    def _convert_value(value):
        if value.lower() == "true":
            return True
        if value.lower() == "false":
            return False
        return value

    if len(ctx.args) % 2 == 1:
        raise TAFError(
            "Custom parameters invalid. Check if there are spaces around each parameter/value"
        )

    custom = (
        {
            ctx.args[i][2:]: _convert_value(ctx.args[i + 1])
            for i in range(0, len(ctx.args), 2)
        }
        if len(ctx.args)
        else {}
    )
    return custom



import json
import os
import fnmatch
import pygit2

from typing import Any, Callable, Dict, List, Optional, Union
from collections import defaultdict
from contextlib import contextmanager
from pathlib import Path
from taf.tuf.storage import GitStorageBackend
from taf.git import GitRepository
from taf.tuf.repository import (
    METADATA_DIRECTORY_NAME,
    MetadataRepository as TUFRepository,
    get_role_metadata_path,
    get_target_path,
)
from taf.constants import INFO_JSON_PATH


class AuthenticationRepository(GitRepository):

    LAST_VALIDATED_FILENAME = "last_validated_commit"
    LAST_VALIDATED_KEY = "last_validated_commit"
    TEST_REPO_FLAG_FILE = "test-auth-repo"
    SCRIPTS_PATH = "scripts"

    _conf_dir = None
    _dependencies: Dict = {}

    def __init__(
        self,
        library_dir: Optional[Union[str, Path]] = None,
        name: Optional[str] = None,
        urls: Optional[List[str]] = None,
        custom: Optional[Dict] = None,
        default_branch: Optional[str] = None,
        allow_unsafe: Optional[bool] = False,
        conf_directory_root: Optional[str] = None,
        out_of_band_authentication: Optional[str] = None,
        path: Optional[Union[Path, str]] = None,
        alias: Optional[str] = None,
        *args,
        **kwargs,
    ):
        """
        Args:
          library_dir (Path or str): path to the library root. This is a directory which contains other repositories,
          whose full path is determined by appending their names to the library dir. A repository can be
          instantiated by specifying library dir and name, or by the full path.
          name (str): repository's name, which is appended to the library dir to form the full path.
          Must be in the namespace/name format. If library_dir is specified, name must be specified too.
          path (Path): repository's full filesystem path, which can be specified instead of name and library dir
          urls (list): repository's urls
          custom (dict): a dictionary containing other data
          default_branch (str): repository's default branch, automatically determined if not specified
          out_of_band_authentication (str): manually specified initial commit
          alias: Repository's alias, which will be used in logging statements to reference it
        """
        super().__init__(
            library_dir,
            name,
            urls,
            custom,
            default_branch,
            allow_unsafe,
            path,
            alias,
            *args,
            **kwargs,
        )

        if conf_directory_root is None:
            conf_directory_root_path = Path(self.path).parent
        else:
            conf_directory_root_path = Path(conf_directory_root)

        self.conf_directory_root = conf_directory_root_path.resolve()
        self.out_of_band_authentication = out_of_band_authentication
        self._storage = GitStorageBackend()
        self._tuf_repository = TUFRepository(self.path, storage=self._storage)

    def __getattr__(self, item):
        """Delegate attribute lookup to TUFRepository instance"""
        if item in self.__dict__:
            return self.__dict__[item]
        try:
            # First try to get attribute from super class (GitRepository)
            return super().__getattribute__(item)
        except AttributeError:
            # If not found, delegate to TUFRepository
            return getattr(self._tuf_repository, item)

    def __dir__(self):
        """Return list of attributes available on this object, including those
        from TUFRepository."""
        return super().__dir__() + dir(self._tuf_repository)

    # TODO rework conf_dir

    def to_json_dict(self) -> Dict:
        """Returns a dictionary mapping all attributes to their values"""
        data = super().to_json_dict()
        data.update(
            {
                "conf_directory_root": str(self.conf_directory_root),
                "out_of_band_authentication": self.out_of_band_authentication,
                "dependencies": self.dependencies,
            }
        )
        return data

    @property
    def conf_dir(self) -> str:
        """
        Returns location of the directory which stores the authentication repository's
        configuration files. That is, the last validated commit.
        Create the directory if it does not exist.
        """
        # the repository's name consists of the namespace and name (namespace/name)
        # the configuration directory should be _name
        if self._conf_dir is None:
            last_dir = os.path.basename(os.path.normpath(self.path))
            conf_path = self.conf_directory_root / f"_{last_dir}"
            conf_path.mkdir(parents=True, exist_ok=True)
            self._conf_dir = str(conf_path)
        return self._conf_dir

    @property
    def certs_dir(self):
        certs_dir = self.path / "certs"
        certs_dir.mkdir(parents=True, exist_ok=True)
        return str(certs_dir)

    @property
    def dependencies(self) -> Dict:
        return self._dependencies

    @dependencies.setter
    def dependencies(self, value):
        self._dependencies = value

    @property
    def is_test_repo(self, branch: Optional[str] = None) -> bool:
        try:
            if branch is not None:
                commit = self.top_commit_of_branch(branch)
            else:
                commit = self.head_commit_sha()
            targets_data = self.get_metadata("targets", commit)
            if targets_data is None:
                return False
            return self.TEST_REPO_FLAG_FILE in targets_data["signed"]["targets"]
        except Exception as e:
            self._log_debug(f"Could not get targets.json metadata: {e}")
            return False

    @property
    def last_validated_commit(self) -> Optional[str]:
        """
        Last validated commit across the entire set of repositories, including authentication and target repositories.
        It is only validated if the update process does not skip any repository
        """
        try:
            if self.last_validated_data is not None:
                return self.last_validated_data[self.LAST_VALIDATED_KEY]
        except KeyError:
            return None
        return None

    @property
    def last_validated_data(self) -> Optional[dict]:
        """
        A dictionary containing the last validated commits for each repository, including both target repositories
        and the authentication repository. It also includes the last validated commit for when all repositories
        were simultaneously updated.
        """
        last_validated_data = {}
        last_validated_path = Path(self.conf_dir, self.LAST_VALIDATED_FILENAME)
        if last_validated_path.is_file():
            data = last_validated_path.read_text().strip()
            try:
                last_validated_data = json.loads(data)
            except json.decoder.JSONDecodeError:
                if data:
                    last_validated_data = {self.name: data}

        return last_validated_data

    @property
    def log_prefix(self) -> str:
        if self.alias:
            return f"{self.alias}: "
        return f"Auth repo {self.name}: "

    def commit_and_push(
        self,
        commit_msg: Optional[str] = None,
        push: Optional[bool] = True,
        commit: Optional[bool] = True,
    ) -> None:

        if commit:
            if commit_msg is None:
                commit_msg = input("\nEnter commit message and press ENTER\n\n")
            new_commit = self.commit(commit_msg)

            if new_commit and push:
                push_successful = self.push()
                if push_successful:
                    current_branch = self.get_current_branch()
                    if current_branch == self.default_branch:
                        self.set_last_validated_of_repo(
                            self.name, new_commit, set_last_validated_commit=True
                        )
                        self._log_notice(
                            f"Updated last_validated_commit to {new_commit}"
                        )
                    else:
                        self._log_debug(
                            "Not pushing to the default branch, skipping last_validated_commit update."
                        )

    def get_target(self, target_name, commit=None, safely=True) -> Optional[Dict]:
        if commit is None:
            commit = self.head_commit_sha()
        if commit is None:
            return None
        target_path = get_target_path(target_name)
        if safely:
            return self.safely_get_json(commit, target_path)
        else:
            return self.get_json(commit, target_path)

    def get_metadata(
        self, role: str, commit: Optional[str] = None, safely: bool = True
    ) -> Optional[Dict]:
        if commit is None:
            commit = self.head_commit_sha()
        if commit is None:
            return None
        metadata_path = get_role_metadata_path(role)
        if safely:
            return self.safely_get_json(commit, metadata_path)
        else:
            return self.get_json(commit, metadata_path)

    def get_info_json(
        self, commit: Optional[str] = None, safely: bool = True
    ) -> Optional[Dict]:
        head_commit = commit or self.head_commit_sha()
        if head_commit is None:
            return None
        if safely:
            return self.safely_get_json(head_commit, INFO_JSON_PATH)
        else:
            return self.get_json(head_commit, INFO_JSON_PATH)

    def get_metadata_path(self, role):
        return self.path / METADATA_DIRECTORY_NAME / f"{role}.json"

    def get_role_repositories(self, role, parent_role=None):
        """Get repositories of the given role

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)
        - parent_role(str): Name of the parent role of the delegated role. If not specified,
                            it will be set automatically, but this might be slow if there
                            are many delegations.

        Returns:
        Repositories' path from repositories.json that matches given role paths

        Raises:
        - securesystemslib.exceptions.FormatError: If the arguments are improperly formatted.
        - securesystemslib.exceptions.UnknownRoleError: If 'rolename' has not been delegated by this
        """
        if self.is_bare_repository:
            # raise an error for now
            # once we have an ergonomic way to get repositories from a bare repository, remove the error
            raise Exception(
                "Getting role repositories from a bare repository is not yet supported."
            )

        role_paths = self._tuf_repository.get_role_paths(role)

        target_repositories = self._get_target_repositories_from_disk()
        return [
            repo
            for repo in target_repositories
            if any([fnmatch.fnmatch(repo, path) for path in role_paths])
        ]

    def is_commit_authenticated(self, target_name: str, commit: str) -> bool:
        """Checks if passed commit is ever authenticated for given target name."""
        for auth_commit in self.all_commits_on_branch(reverse=False):
            target = self.get_target(target_name, auth_commit)
            if target is None:
                continue
            try:
                if target["commit"] == commit:
                    return True
            except TypeError:
                continue
        return False

    @contextmanager
    def repository_at_revision(self, commit: str):
        """
        Context manager that enables reading metadata from an older commit.
        This should be used in combination with the Git storage backend.
        """
        self._storage.commit = commit
        yield
        self._storage.commit = None

    def set_last_validated_data(
        self,
        last_validated_data: dict,
        set_last_validated_commit: Optional[bool] = True,
    ):
        """
        Set the last validated data of the authentication repository.
        In case of a partial update (update run with the --exclude-target option),
        update last validated commits of target repositories that were updated
        """
        if set_last_validated_commit:
            last_validated_data[self.LAST_VALIDATED_KEY] = last_validated_data[
                self.name
            ]
        last_data_str = json.dumps(last_validated_data, indent=4)
        self._log_debug(f"setting last validated data to: {last_data_str}")
        Path(self.conf_dir, self.LAST_VALIDATED_FILENAME).write_text(last_data_str)

    def set_last_validated_of_repo(
        self,
        repo_name: str,
        commit: str,
        set_last_validated_commit: Optional[bool] = True,
    ):
        last_validated_data = self.last_validated_data or {}
        last_validated_data[repo_name] = commit
        last_validated_data[self.LAST_VALIDATED_KEY] = commit
        last_data_str = json.dumps(last_validated_data, indent=4)
        self._log_debug(f"setting last validated data to: {last_data_str}")
        if set_last_validated_commit and self.name == repo_name:
            last_validated_data[self.LAST_VALIDATED_KEY] = last_validated_data[
                self.name
            ]
        Path(self.conf_dir, self.LAST_VALIDATED_FILENAME).write_text(last_data_str)

    def auth_repo_commits_after_repos_last_validated(
        self, target_repos: List, last_validated_data
    ) -> List[str]:
        """
        Traverses the commit history from the most recent commit back to the oldest last validated commit
        of the target repositories. It then quantifies how many of these commits are related to each target repository.

        Returns:
            tuple:
                - List[str]: A list of commit hashes from the oldest last validated commit to the newest commit
                in the authentication repository. This list provides a sequential history of commits affecting
                the target repositories.
        """
        last_validated_target_commits = defaultdict(list)
        for repo in target_repos:
            last_validated_commit = last_validated_data[repo.name]
            last_validated_target_commits[last_validated_commit].append(repo)

        repo = self.pygit_repo

        default_branch = repo.lookup_branch(self.default_branch)
        top_commit = default_branch.peel()

        walker = repo.walk(top_commit.id, pygit2.GIT_SORT_TOPOLOGICAL)

        traversed_commits = []
        for commit in walker:
            commit_id = str(commit.id)
            if commit_id in last_validated_target_commits:
                last_validated_target_commits.pop(commit_id)

            traversed_commits.append(commit_id)
            if not len(last_validated_target_commits):
                break
        traversed_commits.reverse()
        return traversed_commits

    def targets_data_by_auth_commits(
        self,
        commits: List[str],
        target_repos: Optional[List[str]] = None,
        custom_fns: Optional[Dict[str, Callable]] = None,
        default_branch: Optional[str] = None,
        excluded_target_globs: Optional[List[str]] = None,
        last_commits_per_repos: Optional[Dict[str, List]] = None,
    ) -> Dict[str, Dict[str, Dict[str, Any]]]:
        """
        Return a dictionary where each target repository has associated authentication commits,
        and for each authentication commit, there's a dictionary of the branch, commit and custom data.

        {
            'target_repo1': {
                'auth_commit1': {'branch': 'branch1', 'commit': 'commit1', 'custom': {}},
                'auth_commit2': {'branch': 'branch1', 'commit': 'commit2', 'custom': {}},
                ...
            },
            'target_repo2': {
                ...
            },
            ...
        }

        """
        repositories_commits: Dict[str, Dict[str, Dict[str, Any]]] = {}
        targets = self.targets_at_revisions(
            commits,
            target_repos=target_repos,
            default_branch=default_branch,
            last_commits_per_repos=last_commits_per_repos,
        )
        excluded_target_globs = excluded_target_globs or []
        for commit in commits:
            for target_path, target_data in targets[commit].items():
                if any(
                    fnmatch.fnmatch(target_path, excluded_target_glob)
                    for excluded_target_glob in excluded_target_globs
                ):
                    continue

                target_branch = target_data.get("branch")
                target_commit = target_data.get("commit")
                target_data.setdefault("custom", {})
                if custom_fns is not None and target_path in custom_fns:
                    target_data["custom"].update(custom_fns[target_path](target_commit))

                repositories_commits.setdefault(target_path, {})[commit] = {
                    "branch": target_branch,
                    "commit": target_commit,
                    "custom": target_data.get("custom"),
                }

        self._log_debug(
            f"new commits per repositories according to target files: {repositories_commits}"
        )
        return repositories_commits

    def sorted_commits_and_branches_per_repositories(
        self,
        commits: List[str],
        target_repos: Optional[List[str]] = None,
        custom_fns: Optional[Dict[str, Callable]] = None,
        default_branch: Optional[str] = None,
        excluded_target_globs: Optional[List[str]] = None,
    ) -> Dict[str, Dict[str, List[Dict[str, Any]]]]:
        """Return a dictionary consisting of branches and commits belonging
        to it for every target repository:
        {
            target_repo1: {
                branch1: [
                    {"commit": commit1, "custom": {...}, "auth_commit": auth_commit1},
                    {"commit": commit2, "custom": {...}, "auth_commit": auth_commit2},
                    {"commit": commit3, "custom": {...}, "auth_commit": auth_commit3}
                ],
                branch2: [
                    {"commit": commit4, "custom": {...}, "auth_commit": auth_commit4},
                    {"commit": commit5, "custom": {...}, "auth_commit": auth_commit5}
                ]
            },
            target_repo2: {
                branch1: [
                    {"commit": commit6, "custom": {...}, "auth_commit": auth_commit6}
                ],
                branch2: [
                    {"commit": commit7, "custom": {...}, "auth_commit": auth_commit7}
                ]
            }
        }
        Keep in mind that targets metadata
        file is not updated everytime something is committed to the authentication repo.
        """
        repositories_commits: Dict = defaultdict(dict)
        targets = self.targets_at_revisions(
            commits, target_repos=target_repos, default_branch=default_branch
        )
        previous_commits: Dict = {}
        skipped_targets = []
        excluded_target_globs = excluded_target_globs or []
        for commit in commits:
            for target_path, target_data in targets[commit].items():
                if target_path in skipped_targets:
                    continue
                if any(
                    fnmatch.fnmatch(target_path, excluded_target_glob)
                    for excluded_target_glob in excluded_target_globs
                ):
                    skipped_targets.append(target_path)
                    continue
                target_branch = target_data.get("branch")
                target_commit = target_data.get("commit")
                previous_data = previous_commits.get(target_path)
                target_data.setdefault("custom", {})
                if (
                    previous_data is None
                    or (target_commit, target_branch) != previous_data
                ):
                    if custom_fns is not None and target_path in custom_fns:
                        target_data["custom"].update(
                            custom_fns[target_path](target_commit)
                        )

                    repositories_commits[target_path].setdefault(
                        target_branch, []
                    ).append(
                        {
                            "commit": target_commit,
                            "custom": target_data.get("custom"),
                            "auth_commit": commit,
                        }
                    )
                previous_commits[target_path] = (target_commit, target_branch)
        self._log_debug(
            f"new commits per repositories according to target files: {repositories_commits}"
        )
        return repositories_commits

    def targets_at_revisions(
        self,
        commits,
        target_repos=None,
        default_branch=None,
        last_commits_per_repos=None,
    ):
        targets = defaultdict(dict)
        if default_branch is None:
            default_branch = self.default_branch
        previous_metadata = []
        new_files = []
        repos_to_skip = []
        for commit in reversed(commits):
            # repositories.json might not exit, if the current commit is
            # the initial commit
            repositories_at_revision = self.safely_get_json(
                commit, get_target_path("repositories.json")
            )
            if repositories_at_revision is None:
                continue
            repositories_at_revision = repositories_at_revision["repositories"]

            current_metadata = self.list_files_at_revision(
                commit, METADATA_DIRECTORY_NAME
            )
            new_files = [
                metadata_path
                for metadata_path in current_metadata
                if metadata_path not in previous_metadata
            ]
            previous_metadata = current_metadata
            if len(new_files):
                with self.repository_at_revision(commit):
                    roles_at_revision = self.get_all_targets_roles()

            for role_name in roles_at_revision:
                # targets metadata files corresponding to the found roles must exist
                targets_at_revision = self.safely_get_json(
                    commit, get_role_metadata_path(role_name)
                )
                if targets_at_revision is None:
                    continue
                targets_at_revision = targets_at_revision["signed"]["targets"]

                for target_path in targets_at_revision:
                    # if there are older auth repo commits corresponding to repositories
                    # that were not validated in the one or more previous updates
                    # skip the ones that were validated more recently
                    # when the last validated commit of a repo is reached
                    # the repo is added to the repos_to_skip list
                    if target_path in repos_to_skip:
                        continue
                    if (
                        last_commits_per_repos
                        and last_commits_per_repos.get(target_path) == commit
                    ):
                        repos_to_skip.append(target_path)
                    if target_path not in repositories_at_revision:
                        # we only care about repositories
                        continue
                    if target_repos and target_path not in target_repos:
                        # if specific target repositories are specified, skip all other
                        # repositories
                        continue
                    target_content = self.safely_get_json(
                        commit, get_target_path(target_path)
                    )
                    if target_content is not None:
                        target_commit = target_content.pop("commit")
                        target_branch = target_content.pop("branch", default_branch)
                        targets[commit][target_path] = {
                            "branch": target_branch,
                            "commit": target_commit,
                            "custom": target_content,
                        }
        return targets

    def _get_target_repositories_from_disk(self):
        """
        Read repositories.json from disk and return the list of target repositories
        """
        repositories_path = self.targets_path / "repositories.json"
        if repositories_path.exists():
            repositories = repositories_path.read_text()
            repositories = json.loads(repositories)["repositories"]
            return [str(Path(target_path).as_posix()) for target_path in repositories]



from contextlib import contextmanager
import io
from pathlib import Path
from typing import IO, Dict, Optional
import pygit2
from taf.constants import METADATA_DIRECTORY_NAME
from taf.exceptions import GitError, TAFError
from taf.git import GitRepository

from securesystemslib.storage import FilesystemBackend

from securesystemslib.exceptions import StorageError

git_repos_cache: Dict[str, GitRepository] = {}


def is_subpath(path, potential_subpath):
    path = Path(path).resolve()
    potential_subpath = Path(potential_subpath).resolve()

    try:
        potential_subpath.relative_to(path)
        return True
    except ValueError:
        return False


def find_git_repository(inner_path):
    """
    Instantiates a git repository based on a path
    that is expected to be inside that repository.
    Enables smoother integration with TUF's default
    FilesystemBackend implementation
    """
    for path in list(git_repos_cache.keys()):
        if is_subpath(inner_path, path):
            return git_repos_cache[path]
    repo_path = pygit2.discover_repository(inner_path)
    repo = None
    if not repo_path:
        # could be a bare repository
        repo_path = str(inner_path).split(METADATA_DIRECTORY_NAME)[0]
        if repo_path:
            try:
                pygit2.Repository(repo_path)
            except Exception:
                return None
            else:
                repo = GitRepository(path=repo_path)
    else:
        repo_path = Path(repo_path).parent
        repo = GitRepository(path=repo_path)

    if repo:
        git_repos_cache[repo.path] = repo
    return repo


class GitStorageBackend(FilesystemBackend):
    """
    Storage backend implemnantation that loads data
    from a specific commit, if it is specified,
    or from the filesystem, if the commit is None
    Extends TUF's FilesystemBackend.
    """

    commit: Optional[str] = None

    def __new__(cls, *args, **kwargs):
        # Bypass singleton
        # This is necessary in order to use this within the context of
        # parallel update of multiple repositories
        return super(FilesystemBackend, cls).__new__(cls, *args, **kwargs)

    @contextmanager
    def get(self, filepath: str):
        # If the commit is specified, read from Git.
        # If it is not specified, read from the filesystem.
        if self.commit is None:
            with super().get(filepath=filepath) as value_from_base:
                yield value_from_base
        else:
            try:
                repo = find_git_repository(filepath)
                file_path = Path(filepath)
                relative_path = file_path.relative_to(repo.path)
                data = repo.get_file(self.commit, relative_path).encode()
                yield io.BytesIO(data)
            except GitError as e:
                raise StorageError(e)
            except TAFError as e:
                raise StorageError(e)

    def getsize(self, filepath: str) -> int:
        # Get size of a file after reading it from Git or the filesystem.
        # If the commit is specified, read from Git.
        # If it is not specified, read from the filesystem.
        if self.commit is None:
            return super().getsize(filepath=filepath)
        try:
            repo = find_git_repository(filepath)
            file_path = Path(filepath)
            relative_path = file_path.relative_to(repo.path)
            data = repo.get_file(self.commit, relative_path).encode()
            return len(data)
        except GitError as e:
            raise StorageError(e)
        except TAFError as e:
            raise StorageError(e)

    def put(self, fileobj: IO, filepath: str, restrict: Optional[bool] = False) -> None:
        # Write the file to the filesystem.
        # Raise an error if the repository is a bare repository.
        repo_path = pygit2.discover_repository(filepath)
        if repo_path:
            repo = find_git_repository(filepath)
            if repo.is_bare_repository:
                raise TAFError(
                    f"Cannot write to {filepath}. Repository is a bare repository"
                )
        super().put(fileobj, filepath, restrict)



import click
from taf.tools.cli.lazy_group import LazyGroup


@click.group(cls=LazyGroup, lazy_subcommands={
    "repo": "taf.tools.repo.attach_to_group"
})
@click.version_option()
def olc():
    """OLC Command Line Interface"""
    pass


if __name__ == '__main__':
    olc()



import click
from taf.api.dependencies import add_dependency, remove_dependency
from taf.exceptions import TAFError
from taf.tools.cli import catch_cli_exception, find_repository, process_custom_command_line_args


def add_dependency_command():
    @click.command(context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ), help="""Add a dependency (an authentication repository) to dependencies.json or update it if it was already added to this file.
        Update and sign targets metadata, snapshot and timestamp using yubikeys or keys loaded from the specified keystore location.
        Information that is added to dependencies.json includes out-of-band authentication commit and name
        of the branch which contains that commit. This out-of-band authentication commit represents a commit including and following
        which state of the authentication repository is expected to be valid at every commit. If the dependency does not exist
        on the disk, this command will still update dependencies.json, but will not be able to validate branch and out-of-band
        commit values, so it is strongly recommended to run the updater first and clone/update and validate the dependency.

        All additional information that should be saved as the dependency's custom content in `dependencies.json`
        is specified by providing additional options. Here is an example:

        `taf dependencies add --path auth-path dependency-name --branch-name main --out-of-band-commit d4d768da4e8f74f54c644923b7ed0e19a0faf3c5 --custom-property some-value --keystore keystore-path`

        or

        `taf dependencies add dependency-name --branch-name main --out-of-band-commit d4d768da4e8f74f54c644923b7ed0e19a0faf3c5 --custom-property some-value --keystore keystore-path`

        if inside an authentication repository

        In this case, custom-property: some-value will be added to the custom part of the dependency dependencies.json.

        If branch-name and out-of-band-commit are omitted, the default branch and its first commit will be written to dependencies.json.

        Dependency does not have to exist on the filesystem, but if it does, provided branch name and out-of-band commit sha
        will be validated, so it is recommended to run the updater first and update/clone and validate the dependency first.
        If the dependency's full path is not provided, it is expected to be located in the same
        library root directory as the authentication repository, in a directory whose name corresponds to its name.
        If dependency's parent authentication repository's path is `E:\\examples\\root\\namespace\\auth`, and the dependency's namespace prefixed name is
        `namespace1\\auth`, the target's path will be set to `E:\\examples\\root\\namespace1\\auth`.""")
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.argument("dependency_name")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--branch-name", default=None, help="Name of the branch which contains the out-of-band commit")
    @click.option("--dependency-url", default=None, help="URL from which the dependency should be cloned if not already on disk")
    @click.option("--out-of-band-commit", default=None, help="Out-of-band commit SHA")
    @click.option("--dependency-path", default=None, help="Dependency's filesystem path")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not located inside the keystore directory")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be committed automatically")
    @click.pass_context
    def add(ctx, dependency_name, path, branch_name, dependency_url, out_of_band_commit, dependency_path, keystore, prompt_for_keys, no_commit):
        custom = process_custom_command_line_args(ctx)
        add_dependency(
            path=path,
            dependency_name=dependency_name,
            branch_name=branch_name,
            dependency_url=dependency_url,
            out_of_band_commit=out_of_band_commit,
            dependency_path=dependency_path,
            keystore=keystore,
            custom=custom,
            prompt_for_keys=prompt_for_keys,
            commit=not no_commit,
        )
    return add


def remove_dependency_command():
    @click.command(help="""Remove a dependency from dependencies.json.
        Update and sign targets metadata, snapshot and timestamp using yubikeys or keys loaded from the specified keystore location.

        `taf dependencies remove --path auth-path dependency-name --keystore keystore-path`

        or

        `taf dependencies remove dependency-name --keystore keystore-path`

        if inside an authentication repository""")
    @find_repository
    @catch_cli_exception(handle=TAFError)
    @click.argument("dependency-name")
    @click.option("--path", default=".", help="Authentication repository's location. If not specified, set to the current directory")
    @click.option("--keystore", default=None, help="Location of the keystore files")
    @click.option("--prompt-for-keys", is_flag=True, default=False, help="Whether to ask the user to enter their key if not located inside the keystore directory")
    @click.option("--no-commit", is_flag=True, default=False, help="Indicates that the changes should not be committed automatically")
    def remove(dependency_name, path, keystore, prompt_for_keys, no_commit):
        remove_dependency(
            path=path,
            dependency_name=dependency_name,
            keystore=keystore,
            prompt_for_keys=prompt_for_keys,
            commit=not no_commit,
        )
    return remove


def attach_to_group(group):
    group.add_command(add_dependency_command(), name='add')
    group.add_command(remove_dependency_command(), name='remove')



import click
from taf.tools.cli.lazy_group import LazyGroup


@click.group(cls=LazyGroup, lazy_subcommands={
    "dependencies": "taf.tools.dependencies.attach_to_group",
    "keystore": "taf.tools.keystore.attach_to_group",
    "conf": "taf.tools.conf.attach_to_group",
    "repo": "taf.tools.repo.attach_to_group",
    "targets": "taf.tools.targets.attach_to_group",
    "metadata": "taf.tools.metadata.attach_to_group",
    "roles": "taf.tools.roles.attach_to_group",
    "yubikey": "taf.tools.yubikey.attach_to_group",
})
@click.version_option()
def taf():
    """TAF Command Line Interface"""
    pass


if __name__ == '__main__':
    taf()



MESSAGES = {
    "git-commit": {
        "create-repo": "Initialize repository",
        "add-dependency": "Add new dependency {dependency_name}",
        "remove-dependency": "Remove dependency {dependency_name}",
        "add-target": "Add new target {target_name}",
        "remove-target": "Remove target {target_name}",
        "remove-from-delegated-paths": "Remove {target_name} from delegated paths",
        "update-targets": "Sign targets and metadata",
        "update-expiration-dates": "Update expiration date of {roles}",
        "add-role": "Add new role {role}",
        "remove-role": "Remove role {role}",
        "add-role-paths": "Delegate path(s) {paths} to role {role}",
        "add-roles": "Add new roles {roles}",
        "add-signing-key": "Add new signing key to role {role}",
        "remove-role-paths": "Remove delegations {paths} from role {role}",
    }
}


def git_commit_message(key, **kwargs):
    if not len(kwargs):
        return MESSAGES["git-commit"][key]
    return MESSAGES["git-commit"][key].format(**kwargs)
